import unittest
import os
import shutil
import fs
import fs.copy
from fontTools.ufoLib import UFOReader
from defcon import Font
from defcon.test.testTools import (
    getTestFontPath, getTestFontCopyPath, makeTestFontCopy,
    openTestFontAsFileSystem, closeTestFontAsFileSystem,
    tearDownTestFontCopy)


try:
    from plistlib import load, dump, dumps
except ImportError:
    from plistlib import readPlist as load, writePlist as dump


class LayerTest(unittest.TestCase):

    def __init__(self, methodName):
        unittest.TestCase.__init__(self, methodName)
        if not hasattr(self, "assertRaisesRegex"):
            self.assertRaisesRegex = self.assertRaisesRegexp

    def tearDown(self):
        if os.path.exists(getTestFontCopyPath()):
            tearDownTestFontCopy()

    def test_set_parent_data_in_layer(self):
        font = Font(getTestFontPath())
        layers = font.layers
        layer = font.layers[None]
        self.assertEqual(id(layer.getParent()), id(layers))

    def test_defaultLayer(self):
        font = Font(getTestFontPath())
        layers = font.layers
        layer = layers.defaultLayer
        self.assertEqual(layer, layers["public.default"])
        layer = layers["Layer 1"]
        layers.defaultLayer = layer
        self.assertEqual(layer, layers.defaultLayer)

    def test_layerOrder(self):
        font = Font(getTestFontPath())
        layers = font.layers
        self.assertEqual(layers.layerOrder,
                         ["public.default", "public.background", "Layer 1"])
        self.assertEqual(list(reversed(layers.layerOrder)),
                         ["Layer 1", "public.background", "public.default"])

    def test_newLayer(self):
        font = Font(getTestFontPath())
        layers = font.layers
        layer = font.newLayer("Test")
        self.assertTrue(layer.dirty)
        self.assertTrue(layers.dirty)
        self.assertTrue(font.dirty)
        self.assertEqual(
            layers.layerOrder,
            ["public.default", "public.background", "Layer 1", "Test"])

    def test_iter(self):
        font = Font(getTestFontPath())
        layers = font.layers
        self.assertEqual([layer.name for layer in layers],
                         ["public.default", "public.background", "Layer 1"])

    def test_getitem(self):
        font = Font(getTestFontPath())
        layers = font.layers
        self.assertEqual(layers["public.default"].name, "public.default")

    def test_delitem(self):
        font = Font(makeTestFontCopy())
        path = os.path.join(font.path, "glyphs.public.background")
        self.assertTrue(os.path.exists(path))
        layers = font.layers
        del layers["public.background"]
        layers.dirty = True
        self.assertEqual(layers.layerOrder, ["public.default", "Layer 1"])
        self.assertNotIn("public.background", layers)

        self.assertEqual(len(layers), 2)
        with self.assertRaisesRegex(KeyError, "public.background"):
            layers["public.background"]
        font.save()
        path = os.path.join(font.path, "glyphs.public.background")
        self.assertFalse(os.path.exists(path))
        tearDownTestFontCopy()

        font = Font(makeTestFontCopy())
        path = os.path.join(font.path, "glyphs.public.background")
        del font.layers["public.background"]
        layer = font.newLayer("public.background")
        layer.newGlyph("B")
        font.save()
        self.assertFalse(os.path.exists(os.path.join(path, "A_.glif")))
        self.assertTrue(os.path.exists(os.path.join(path, "B_.glif")))

    def test_len(self):
        with Font(getTestFontPath()) as font:
            layers = font.layers
            self.assertEqual(len(layers), 3)

        font = Font()
        layers = font.layers
        self.assertEqual(len(layers), 1)

    def test_contains(self):
        with Font(getTestFontPath()) as font:
            layers = font.layers
            self.assertIn("public.default", layers)
            self.assertNotIn("NotInFont", layers)

    def test_testForExternalChanges(self):
        for ufo in (u"TestExternalEditing.ufo", u"TestExternalEditing.ufoz"):
            path = getTestFontPath(ufo)
            path = makeTestFontCopy(path)
            with Font(path) as font, UFOReader(path) as reader:
                self.assertEqual(font.layers.testForExternalChanges(reader),
                                 {"deleted": [], "added": [], "modified": {},
                                  "defaultLayer": False, "order": False})
            tearDownTestFontCopy(font.path)

    def test_testForExternalChanges_layerinfo(self):
        for ufo in (u"TestExternalEditing.ufo", u"TestExternalEditing.ufoz"):
            # layerinfo.plist
            path = getTestFontPath(ufo)
            path = makeTestFontCopy(path)
            with Font(path) as font:
                data = {"lib": {}}
                data["lib"]["testForExternalChanges.test"] = 1
                fileSystem = openTestFontAsFileSystem(path)
                p = fs.path.join("glyphs", "layerinfo.plist")
                with fileSystem.open(p, mode="wb") as f:
                    dump(data, f)
                closeTestFontAsFileSystem(fileSystem, path)
                with UFOReader(path) as reader:
                    self.assertTrue(
                        font.layers.testForExternalChanges(reader)
                        ["modified"]["public.default"]["info"])
            tearDownTestFontCopy(font.path)

    def test_testForExternalChanges_add_a_layer(self):
        for ufo in (u"TestExternalEditing.ufo", u"TestExternalEditing.ufoz"):
            path = getTestFontPath(ufo)
            path = makeTestFontCopy(path)
            with Font(path) as font:
                fileSystem = openTestFontAsFileSystem(path)
                fs.copy.copy_dir(fileSystem, "glyphs", fileSystem, "glyphs.test")
                with fileSystem.open(u"layercontents.plist", "rb") as f:
                    contents = load(f)
                contents.append(("test", "glyphs.test"))
                with fileSystem.open(u"layercontents.plist", "wb") as f:
                    dump(contents, f)
                closeTestFontAsFileSystem(fileSystem, path)
                with UFOReader(path) as reader:
                    self.assertEqual(font.layers.testForExternalChanges(reader)["added"],
                                     ["test"])
            tearDownTestFontCopy(font.path)

    def test_testForExternalChanges_remove_a_layer(self):
        for ufo in (u"TestExternalEditing.ufo", u"TestExternalEditing.ufoz"):
            path = getTestFontPath(ufo)
            path = makeTestFontCopy(path)
            fileSystem = openTestFontAsFileSystem(path)
            fs.copy.copy_dir(fileSystem, "glyphs", fileSystem, "glyphs.test")
            with fileSystem.open(u"layercontents.plist", "rb") as f:
                contents = load(f)
            contents.append(("test", "glyphs.test"))
            with fileSystem.open(u"layercontents.plist", "wb") as f:
                dump(contents, f)
            closeTestFontAsFileSystem(fileSystem, path)
            font = Font(path)
            with font:
                fileSystem = openTestFontAsFileSystem(path)
                fileSystem.removetree(u"glyphs.test")
                with fileSystem.open(u"layercontents.plist", "rb") as f:
                    contents = load(f)
                del contents[-1]
                with fileSystem.open(u"layercontents.plist", "wb") as f:
                    dump(contents, f)
                closeTestFontAsFileSystem(fileSystem, path)
                with UFOReader(path) as reader:
                    self.assertEqual(font.layers.testForExternalChanges(reader)["deleted"],
                                     ["test"])
            tearDownTestFontCopy(font.path)

    def test_testForExternalChanges_change_layer_order(self):
        for ufo in (u"TestExternalEditing.ufo", u"TestExternalEditing.ufoz"):
            path = getTestFontPath(ufo)
            path = makeTestFontCopy(path)
            fileSystem = openTestFontAsFileSystem(path)
            fs.copy.copy_dir(fileSystem, "glyphs", fileSystem, "glyphs.test")
            with fileSystem.open(u"layercontents.plist", "rb") as f:
                contents = load(f)
            contents.append(("test", "glyphs.test"))
            with fileSystem.open(u"layercontents.plist", "wb") as f:
                dump(contents, f)
            closeTestFontAsFileSystem(fileSystem, path)
            with Font(path) as font:
                fileSystem = openTestFontAsFileSystem(path)
                with fileSystem.open(u"layercontents.plist", "rb") as f:
                    contents = load(f)
                contents.reverse()
                with fileSystem.open(u"layercontents.plist", "wb") as f:
                    dump(contents, f)
                closeTestFontAsFileSystem(fileSystem, path)
                with UFOReader(path) as reader:
                    self.assertEqual(font.layers.testForExternalChanges(reader),
                                     {"deleted": [], "added": [], "modified": {},
                                      "defaultLayer": False, "order": True})
            tearDownTestFontCopy(font.path)

    def test_testForExternalChanges_change_default_layer(self):
        for ufo in (u"TestExternalEditing.ufo", u"TestExternalEditing.ufoz"):
            path = getTestFontPath(ufo)
            path = makeTestFontCopy(path)

            fileSystem = openTestFontAsFileSystem(path)
            fs.copy.copy_dir(fileSystem, "glyphs", fileSystem, "glyphs.test")
            contents = [("foo", "glyphs"), ("test", "glyphs.test")]
            with fileSystem.open(u"layercontents.plist", "wb") as f:
                dump(contents, f)
            closeTestFontAsFileSystem(fileSystem, path)
            with Font(path) as font:
                fileSystem = openTestFontAsFileSystem(path)
                contents = [("foo", "glyphs.test"), ("test", "glyphs")]
                with fileSystem.open(u"layercontents.plist", "wb") as f:
                    dump(contents, f)
                closeTestFontAsFileSystem(fileSystem, path)
                with UFOReader(path) as reader:
                    self.assertEqual(font.layers.testForExternalChanges(reader),
                                     {"deleted": [], "added": [], "modified": {},
                                      "defaultLayer": True, "order": False})
            tearDownTestFontCopy(font.path)

    def test_reloadLayers_layerinfo(self):
        path = getTestFontPath(u"TestExternalEditing.ufo")
        path = makeTestFontCopy(path)
        font = Font(path)
        p = os.path.join(path, "glyphs", "layerinfo.plist")
        data = {"lib": {}}
        data["lib"]["testForExternalChanges.test"] = 1
        with open(p, "wb") as f:
            dump(data, f)
        font.reloadLayers(dict(layers={"public.default": dict(info=True)}))
        self.assertEqual(font.layers["public.default"].lib,
                         {"testForExternalChanges.test": 1})
        tearDownTestFontCopy(font.path)

    def test_reloadLayers_add_a_layer(self):
        path = getTestFontPath(u"TestExternalEditing.ufo")
        path = makeTestFontCopy(path)
        font = Font(path)
        shutil.copytree(os.path.join(path, "glyphs"),
                        os.path.join(path, "glyphs.test"))
        with open(os.path.join(path, "layercontents.plist"), "rb") as f:
            contents = load(f)
        contents.append(("test", "glyphs.test"))
        with open(os.path.join(path, "layercontents.plist"), "wb") as f:
            dump(contents, f)
        font.reloadLayers(dict(layers={"test": {}}))
        self.assertEqual(font.layers.layerOrder, ["public.default", "test"])
        tearDownTestFontCopy(font.path)

    def test_reloadLayers_change_layer_order(self):
        path = getTestFontPath(u"TestExternalEditing.ufo")
        path = makeTestFontCopy(path)
        shutil.copytree(os.path.join(path, "glyphs"),
                        os.path.join(path, "glyphs.test"))
        with open(os.path.join(path, "layercontents.plist"), "rb") as f:
            contents = load(f)
        contents.append(("test", "glyphs.test"))
        with open(os.path.join(path, "layercontents.plist"), "wb") as f:
            dump(contents, f)
        font = Font(path)
        with open(os.path.join(path, "layercontents.plist"), "rb") as f:
            contents = load(f)
        contents.reverse()
        with open(os.path.join(path, "layercontents.plist"), "wb") as f:
            dump(contents, f)
        font.reloadLayers(dict(order=True))
        self.assertEqual(font.layers.layerOrder, ["test", "public.default"])
        tearDownTestFontCopy(font.path)

    def test_reloadLayers_change_default_layer(self):
        path = getTestFontPath(u"TestExternalEditing.ufo")
        path = makeTestFontCopy(path)
        shutil.copytree(os.path.join(path, "glyphs"),
                        os.path.join(path, "glyphs.test"))
        contents = [("foo", "glyphs"), ("test", "glyphs.test")]
        with open(os.path.join(path, "layercontents.plist"), "wb") as f:
            dump(contents, f)
        font = Font(path)
        contents = [("test", "glyphs"), ("foo", "glyphs.test")]
        contents.reverse()
        with open(os.path.join(path, "layercontents.plist"), "wb") as f:
            dump(contents, f)
        font.reloadLayers(dict(default=True))
        self.assertEqual(font.layers.defaultLayer.name, "test")
        tearDownTestFontCopy(font.path)

    def test_layer_info(self):
        # open and change some values
        font = Font(makeTestFontCopy())
        layer = font.layers["Layer 1"]
        self.assertEqual(layer.color,
                         "0.1,0.2,0.3,0.4")
        layer.color = "0.5,0.5,0.5,0.5"
        self.assertEqual(layer.lib,
                         {"com.typesupply.defcon.test": "1 2 3"})
        layer.lib["foo"] = "bar"
        font.save()
        path = font.path

        # reopen and check the changes
        font = Font(path)
        layer = font.layers["Layer 1"]
        self.assertEqual(layer.color,
                         "0.5,0.5,0.5,0.5")
        self.assertEqual(sorted(layer.lib.items()),
                         [("com.typesupply.defcon.test", "1 2 3"),
                          ("foo", "bar")])
        tearDownTestFontCopy()

    def test_name_change(self):
        font = Font(getTestFontPath())
        layers = font.layers
        layer = layers["public.background"]
        layers.dirty = False
        layer.dirty = False
        layer.name = "Name Change Test"
        self.assertEqual(layers.layerOrder,
                         ["public.default", "Name Change Test", "Layer 1"])
        self.assertTrue(layer.dirty)

    def test_rename_default_layer(self):
        # https://github.com/unified-font-object/ufoLib/issues/123
        path = getTestFontCopyPath()
        font = Font()
        font.save(path)
        font.layers.defaultLayer.name = "somethingElse"
        font.save()
        self.assertEqual(Font(path).layers.defaultLayer.name, "somethingElse")


if __name__ == "__main__":
    unittest.main()



from __future__ import unicode_literals
import unittest
import os
import fs
import fs.copy
import fs.path
from fontTools.ufoLib import UFOReader
from defcon import Font
from defcon.objects.imageSet import fileNameValidator
from defcon.test.testTools import (
    getTestFontPath, getTestFontCopyPath, makeTestFontCopy,
    openTestFontAsFileSystem, closeTestFontAsFileSystem,
    tearDownTestFontCopy)

pngSignature = b"\x89PNG\r\n\x1a\n"


class ImageSetTest(unittest.TestCase):

    def __init__(self, methodName):
        unittest.TestCase.__init__(self, methodName)

    def tearDown(self):
        path = getTestFontCopyPath()
        if os.path.exists(path):
            tearDownTestFontCopy()

    def test_read(self):
        path = getTestFontPath()
        font = Font(path)
        self.assertEqual(sorted(font.images.fileNames),
                         ["image 1.png", "image 2.png"])

        data = font.images["image 1.png"]
        p = os.path.join(path, "images", "image 1.png")
        f = open(p, "rb")
        expected = f.read()
        f.close()
        self.assertEqual(data, expected)

        data = font.images["image 2.png"]
        p = os.path.join(path, "images", "image 2.png")
        f = open(p, "rb")
        expected = f.read()
        f.close()
        self.assertEqual(data, expected)

    def test_write(self):
        path = makeTestFontCopy()
        font = Font(path)
        font.images["image 3.png"] = font.images["image 1.png"]
        del font.images["image 1.png"]
        font.save()
        p = os.path.join(path, "images", "image 1.png")
        self.assertFalse(os.path.exists(p))
        p = os.path.join(path, "images", "image 2.png")
        self.assertTrue(os.path.exists(p))
        p = os.path.join(path, "images", "image 3.png")
        self.assertTrue(os.path.exists(p))
        tearDownTestFontCopy()

    def test_save_as(self):
        path = getTestFontPath()
        font = Font(path)
        saveAsPath = getTestFontCopyPath(path)
        font.save(saveAsPath)
        imagesDirectory = os.path.join(saveAsPath, "images")
        self.assertTrue(os.path.exists(imagesDirectory))
        imagePath = os.path.join(imagesDirectory, "image 1.png")
        self.assertTrue(os.path.exists(imagePath))
        imagePath = os.path.join(imagesDirectory, "image 2.png")
        self.assertTrue(os.path.exists(imagePath))
        tearDownTestFontCopy(saveAsPath)

    def test_unreferenced_images(self):
        path = getTestFontPath()
        font = Font(path)
        self.assertEqual(font.images.unreferencedFileNames, ["image 2.png"])

        path = makeTestFontCopy()
        font = Font(path)
        font.save(removeUnreferencedImages=True)
        p = os.path.join(path, "images", "image 1.png")
        self.assertTrue(os.path.exists(p))
        p = os.path.join(path, "images", "image 2.png")
        self.assertFalse(os.path.exists(p))
        tearDownTestFontCopy()

    def test_duplicate_image(self):
        path = getTestFontPath()
        font = Font(path)
        data = font.images["image 1.png"]
        self.assertEqual(font.images.findDuplicateImage(data), "image 1.png")
        imagePath = os.path.join(path, "images", "image 2.png")
        f = open(imagePath, "rb")
        data = f.read()
        f.close()
        self.assertEqual(font.images.findDuplicateImage(data), "image 2.png")

    def test_testExternalChanges_remove_in_memory_and_scan(self):
        for ufo in (u"TestExternalEditing.ufo", u"TestExternalEditing.ufoz"):
            path = getTestFontPath(ufo)
            path = makeTestFontCopy(path)
            with Font(path) as font, UFOReader(path) as reader:
                del font.images["image 1.png"]
                self.assertEqual(font.images.testForExternalChanges(reader),
                                 ([], [], []))
            tearDownTestFontCopy(font.path)

    def test_testExternalChanges_add_in_memory_and_scan(self):
        for ufo in (u"TestExternalEditing.ufo", u"TestExternalEditing.ufoz"):
            path = getTestFontPath(ufo)
            path = makeTestFontCopy(path)
            with Font(path) as font, UFOReader(path) as reader:
                font.images["image 3.png"] = pngSignature + b"blah"
                self.assertEqual(font.images.testForExternalChanges(reader),
                                 ([], [], []))
            tearDownTestFontCopy(font.path)

    def test_testExternalChanges_modify_in_memory_and_scan(self):
        for ufo in (u"TestExternalEditing.ufo", u"TestExternalEditing.ufoz"):
            path = getTestFontPath(ufo)
            path = makeTestFontCopy(path)
            with Font(path) as font, UFOReader(path) as reader:
                font.images["image 1.png"] = pngSignature + b"blah"
                self.assertEqual(font.images.testForExternalChanges(reader),
                                 ([], [], []))
            tearDownTestFontCopy(font.path)

    def test_testExternalChanges_remove_on_disk_and_scan(self):
        for ufo in (u"TestExternalEditing.ufo", u"TestExternalEditing.ufoz"):
            path = getTestFontPath(ufo)
            path = makeTestFontCopy(path)
            with Font(path) as font:
                fileSystem = openTestFontAsFileSystem(font.path)
                fileSystem.remove(fs.path.join("images", "image 1.png"))
                closeTestFontAsFileSystem(fileSystem, font.path)
                with UFOReader(path) as reader:
                    self.assertEqual(font.images.testForExternalChanges(reader),
                                     ([], [], ["image 1.png"]))
            tearDownTestFontCopy(font.path)

    def test_testExternalChanges_add_on_disk_and_scan(self):
        for ufo in (u"TestExternalEditing.ufo", u"TestExternalEditing.ufoz"):
            path = getTestFontPath(ufo)
            path = makeTestFontCopy(path)
            with Font(path) as font:
                fileSystem = openTestFontAsFileSystem(font.path)
                source = fs.path.join("images", "image 1.png")
                dest = fs.path.join("images", "image 3.png")
                fileSystem.copy(source, dest)
                closeTestFontAsFileSystem(fileSystem, font.path)
                with UFOReader(path) as reader:
                    self.assertEqual(font.images.testForExternalChanges(reader),
                                     ([], ["image 3.png"], []))
            tearDownTestFontCopy(font.path)

    def test_testExternalChanges_modify_on_disk_and_scan(self):
        for ufo in (u"TestExternalEditing.ufo", u"TestExternalEditing.ufoz"):
            path = getTestFontPath(ufo)
            path = makeTestFontCopy(path)
            with Font(path) as font:
                font.images["image 1.png"]  # image = font.images["image 1.png"]
                fileSystem = openTestFontAsFileSystem(font.path)
                imagePath = fs.path.join("images", "image 1.png")
                data = fileSystem.readbytes(imagePath)
                fileSystem.writebytes(imagePath, data + b"blah")
                closeTestFontAsFileSystem(fileSystem, font.path)
                with UFOReader(path) as reader:
                    self.assertEqual(font.images.testForExternalChanges(reader),
                                     (["image 1.png"], [], []))
            tearDownTestFontCopy(font.path)

    def test_reloadImages(self):
        path = makeTestFontCopy()
        font = Font(path)
        image = font.images["image 1.png"]
        imagePath = os.path.join(path, "images", "image 1.png")
        newImageData = pngSignature + b"blah"
        f = open(imagePath, "wb")
        f.write(newImageData)
        f.close()
        font.images.reloadImages(["image 1.png"])
        image = font.images["image 1.png"]
        self.assertEqual(image, newImageData)
        tearDownTestFontCopy()

    def test_fileNameValidator(self):
        self.assertTrue(fileNameValidator('a'))
        self.assertTrue(fileNameValidator('A_'))
        self.assertTrue(fileNameValidator('A_E_'))
        self.assertTrue(fileNameValidator('A_e'))
        self.assertTrue(fileNameValidator('ae'))
        self.assertTrue(fileNameValidator('aE_'))
        self.assertTrue(fileNameValidator('a.alt'))
        self.assertTrue(fileNameValidator('A_.alt'))
        self.assertTrue(fileNameValidator('A_.A_lt'))
        self.assertTrue(fileNameValidator('A_.aL_t'))
        self.assertTrue(fileNameValidator('A_.alT_'))
        self.assertTrue(fileNameValidator('T__H_'))
        self.assertTrue(fileNameValidator('T__h'))
        self.assertTrue(fileNameValidator('t_h'))
        self.assertTrue(fileNameValidator('F__F__I_'))
        self.assertTrue(fileNameValidator('f_f_i'))
        self.assertTrue(fileNameValidator('A_acute_V_.swash'))
        self.assertTrue(fileNameValidator('_notdef'))
        self.assertTrue(fileNameValidator('_con'))
        self.assertTrue(fileNameValidator('C_O_N_'))
        self.assertTrue(fileNameValidator('_con.alt'))
        self.assertTrue(fileNameValidator('alt._con'))
        self.assertTrue(fileNameValidator('A_bC_dE_f'))

        self.assertFalse(fileNameValidator(b'A'))
        self.assertFalse(fileNameValidator('A'*256))
        self.assertFalse(fileNameValidator('A'))
        self.assertFalse(fileNameValidator('con'))
        self.assertFalse(fileNameValidator('a/alt'))

if __name__ == "__main__":
    unittest.main()



import unittest
from defcon.objects.anchor import Anchor
from defcon.objects.font import Font
from defcon.test.testTools import NotificationTestObserver


class AnchorTest(unittest.TestCase):

    def __init__(self, methodName):
        unittest.TestCase.__init__(self, methodName)

    def setUp(self):
        self.font = Font()
        self.glyph = self.font.newGlyph("A")
        self.anchor = Anchor()

    def tearDown(self):
        del self.anchor
        del self.glyph
        del self.font

    def test_dirty(self):
        self.assertFalse(self.anchor.dirty)
        notdirty = not self.anchor.dirty
        self.anchor.dirty = notdirty
        self.assertEqual(self.anchor.dirty, notdirty)
        self.anchor.dirty = not notdirty
        self.assertNotEqual(self.anchor.dirty, notdirty)

    def test_getParent(self):
        self.assertIsNone(self.anchor.getParent())
        self.anchor = Anchor(self.glyph)
        self.assertEqual(self.anchor.getParent(), self.glyph)

    def test_font(self):
        self.assertIsNone(self.anchor.font)
        self.anchor = Anchor(self.glyph)
        self.assertEqual(self.anchor.font, self.font)
        with self.assertRaises(AttributeError):
            self.anchor.font = "foo"

    def test_layerSet(self):
        self.assertIsNone(self.anchor.layerSet)
        self.anchor = Anchor(self.glyph)
        self.assertEqual(self.anchor.layerSet, self.glyph.layerSet)
        self.assertIsNotNone(self.anchor.layerSet)
        with self.assertRaises(AttributeError):
            self.anchor.layerSet = "foo"

    def test_layer(self):
        self.assertIsNone(self.anchor.layer)
        self.anchor = Anchor(self.glyph)
        self.assertEqual(self.anchor.layer, self.glyph.layer)
        self.assertIsNotNone(self.anchor.layer)
        with self.assertRaises(AttributeError):
            self.anchor.layer = "foo"

    def test_x(self):
        self.anchor.x = 100
        self.assertEqual(self.anchor.x, 100)
        self.assertTrue(self.anchor.dirty)

    def test_y(self):
        self.anchor.y = 100
        self.assertEqual(self.anchor.y, 100)
        self.assertTrue(self.anchor.dirty)

    def test_name(self):
        self.anchor.name = "foo"
        self.assertEqual(self.anchor.name, "foo")
        self.assertTrue(self.anchor.dirty)
        self.anchor.name = None
        self.assertIsNone(self.anchor.name)
        self.assertTrue(self.anchor.dirty)

    def test_color(self):
        self.anchor.color = "1,1,1,1"
        self.assertEqual(self.anchor.color, "1,1,1,1")
        self.assertTrue(self.anchor.dirty)

    def test_identifiers(self):
        anchor = Anchor(self.glyph)
        self.assertEqual(anchor.identifiers, self.glyph.identifiers)

    def test_identifier(self):
        self.assertIsNone(self.anchor.identifier)
        identifier = self.anchor.generateIdentifier()
        self.assertEqual(identifier, self.anchor.identifier)
        self.assertIsNotNone(self.anchor.identifier)

    def test_identifier_set(self):
        self.assertIsNone(self.anchor.identifier)
        self.anchor.identifier = "foo"
        self.assertEqual(self.anchor.identifier, "foo")
        self.anchor.identifier = "bar"
        self.assertEqual(self.anchor.identifier, "foo")
        self.anchor.identifier = None
        self.assertEqual(self.anchor.identifier, "foo")

    def test_instance(self):
        a = Anchor(anchorDict=dict(x=1, y=2, name="3", identifier="4",
                                   color="1,1,1,1"))
        self.assertEqual((a.x, a.y, a.name, a.identifier, a.color),
                         (1, 2, "3", "4", "1,1,1,1"))

    def test_move(self):
        a = Anchor(anchorDict=dict(x=1, y=2, name="3"))
        a.dirty = False
        self.assertEqual((a.x, a.y), (1, 2))
        a.move((10, 0))
        self.assertTrue(a.dirty)

        a.dirty = False
        self.assertEqual((a.x, a.y), (11, 2))
        a.move((0, -123))
        self.assertTrue(a.dirty)

        a.dirty = False
        self.assertEqual((a.x, a.y), (11, -121))
        a.move((-11, 121))
        self.assertEqual((a.x, a.y), (0, 0))
        self.assertTrue(a.dirty)


class AnchorNotificationTest(unittest.TestCase):

    def __init__(self, methodName):
        unittest.TestCase.__init__(self, methodName)

    def setUp(self):
        self.font = Font()
        self.glyph = self.font.newGlyph("A")
        self.anchor = Anchor(
            self.glyph,
            {"name": "anchor1", "x": 300, "y": 700}
        )
        self.notificationObject = NotificationTestObserver()

    def tearDown(self):
        del self.anchor
        del self.glyph
        del self.font
        del self.notificationObject

    def test_x_changed_notification(self):
        self.anchor.dispatcher.addObserver(
            observer=self.notificationObject,
            methodName="notificationCallback",
            notification="Anchor.XChanged",
            observable=self.anchor
        )
        self.assertEqual(self.notificationObject.stack, [])
        self.anchor.x += 10
        self.assertEqual(
            self.notificationObject.stack[-1],
            ("Anchor.XChanged", "anchor1")
        )

    def test_y_changed_notification(self):
        self.anchor.dispatcher.addObserver(
            observer=self.notificationObject,
            methodName="notificationCallback",
            notification="Anchor.YChanged",
            observable=self.anchor
        )
        self.assertEqual(self.notificationObject.stack, [])
        self.anchor.y += 10
        self.assertEqual(
            self.notificationObject.stack[-1],
            ("Anchor.YChanged", "anchor1")
        )

    def test_name_notification(self):
        self.anchor.dispatcher.addObserver(
            observer=self.notificationObject,
            methodName="notificationCallback",
            notification="Anchor.NameChanged",
            observable=self.anchor
        )
        self.assertEqual(self.notificationObject.stack, [])
        self.anchor.name += "_suffix"
        self.assertEqual(
            self.notificationObject.stack[-1],
            ("Anchor.NameChanged", "anchor1_suffix")
        )

    def test_color_notification(self):
        self.anchor.dispatcher.addObserver(
            observer=self.notificationObject,
            methodName="notificationCallback",
            notification="Anchor.ColorChanged",
            observable=self.anchor
        )
        self.assertEqual(self.notificationObject.stack, [])
        self.anchor.color = "1,1,1,1"
        self.assertEqual(
            self.notificationObject.stack[-1],
            ("Anchor.ColorChanged", "anchor1")
        )

    def test_identifier_notification(self):
        self.anchor.dispatcher.addObserver(
            observer=self.notificationObject,
            methodName="notificationCallback",
            notification="Anchor.IdentifierChanged",
            observable=self.anchor
        )
        self.assertEqual(self.notificationObject.stack, [])
        self.anchor.identifier = "anchor1_identifier"
        self.assertEqual(
            self.notificationObject.stack[-1],
            ("Anchor.IdentifierChanged", "anchor1")
        )

    def test_endSelfNotificationObservation(self):
        self.assertIsNotNone(self.anchor.dispatcher)
        self.assertIsNotNone(self.anchor.font)
        self.assertIsNotNone(self.anchor.layerSet)
        self.assertIsNotNone(self.anchor.layer)
        self.assertIsNotNone(self.anchor.glyph)

        self.anchor.endSelfNotificationObservation()

        self.assertIsNone(self.anchor.dispatcher)
        self.assertIsNone(self.anchor.font)
        self.assertIsNone(self.anchor.layerSet)
        self.assertIsNone(self.anchor.layer)
        self.assertIsNone(self.anchor.glyph)


if __name__ == "__main__":
    unittest.main()



from __future__ import absolute_import
from fontTools.pens.areaPen import AreaPen
from fontTools.pens.boundsPen import ControlBoundsPen, BoundsPen
from fontTools.misc.arrayTools import unionRect

# ------
# Groups
# ------

def kerningSide1GroupsRepresentationFactory(groups):
    return _gatherGroupsWithPrefix(groups, "public.kern1.")

def kerningSide2GroupsRepresentationFactory(groups):
    return _gatherGroupsWithPrefix(groups, "public.kern2.")

def _gatherGroupsWithPrefix(groups, prefix):
    found = {}
    for name, glyphNames in groups.items():
        if name.startswith(prefix):
            found[name] = glyphNames
    return found

def glyphToKerningSide1GroupsRepresentationFactory(groups):
    groups = groups.getRepresentation("defcon.groups.kerningSide1Groups")
    return _makeGlyphToGroupMapping(groups)

def glyphToKerningSide2GroupsRepresentationFactory(groups):
    groups = groups.getRepresentation("defcon.groups.kerningSide2Groups")
    return _makeGlyphToGroupMapping(groups)

def _makeGlyphToGroupMapping(groups):
    glyphToGroup = {}
    for groupName, glyphNames in groups.items():
        for glyphName in glyphNames:
            glyphToGroup[glyphName] = groupName
    return glyphToGroup

# -----
# Glyph
# -----

def glyphAreaRepresentationFactory(glyph):
    pen = AreaPen(glyph.layer)
    glyph.draw(pen)
    return abs(pen.value)

# -------
# Contour
# -------

# bounds

def contourBoundsRepresentationFactory(obj):
    pen = BoundsPen(None)
    obj.draw(pen)
    return pen.bounds

def contourControlPointBoundsRepresentationFactory(obj):
    pen = ControlBoundsPen(None)
    obj.draw(pen)
    return pen.bounds

# area

def contourAreaRepresentationFactory(contour):
    pen = AreaPen()
    pen._endPath = pen._closePath
    contour.draw(pen)
    return pen.value

# flattened

def contourFlattenedRepresentationFactory(contour, approximateSegmentLength=5, segmentLines=False):
    from fontPens.flattenPen import FlattenPen
    from defcon.objects.glyph import Glyph
    contourClass = contour.__class__
    glyph = Glyph(contourClass=contourClass)
    outputPen = glyph.getPen()
    flattenPen = FlattenPen(outputPen, approximateSegmentLength=approximateSegmentLength, segmentLines=segmentLines)
    contour.draw(flattenPen)
    output = glyph[0]
    return output

# ---------
# Component
# ---------

# bounds

def componentBoundsRepresentationFactory(obj):
    pen = BoundsPen(obj.layer)
    obj.draw(pen)
    return pen.bounds

def componentPointBoundsRepresentationFactory(obj):
    pen = ControlBoundsPen(obj.layer)
    obj.draw(pen)
    return pen.bounds



import unittest
from defcon import Guideline


class GuidelineTest(unittest.TestCase):

    def __init__(self, methodName):
        unittest.TestCase.__init__(self, methodName)

    def test_x(self):
        g = Guideline()
        g.x = 100
        self.assertEqual(g.x, 100)
        self.assertTrue(g.dirty)
        g.x = None
        self.assertIsNone(g.x)
        self.assertTrue(g.dirty)

    def test_y(self):
        g = Guideline()
        g.y = 100
        self.assertEqual(g.y, 100)
        self.assertTrue(g.dirty)
        g.y = None
        self.assertIsNone(g.y)
        self.assertTrue(g.dirty)

    def test_angle(self):
        g = Guideline()
        g.angle = 100
        self.assertEqual(g.angle, 100)
        self.assertTrue(g.dirty)
        g.angle = None
        self.assertIsNone(g.angle)
        self.assertTrue(g.dirty)

    def test_name(self):
        g = Guideline()
        g.name = "foo"
        self.assertEqual(g.name, "foo")
        self.assertEqual(g.keys(), {"name"})
        self.assertTrue(g.dirty)
        g.name = None
        self.assertIsNone(g.name)
        self.assertEqual(g.keys(), set())
        self.assertTrue(g.dirty)

    def test_color(self):
        g = Guideline()
        g.color = "1,1,1,1"
        self.assertEqual(g.color, "1,1,1,1")
        self.assertTrue(g.dirty)
        g.color = None
        self.assertIsNone(g.color)
        self.assertTrue(g.dirty)

    def test_identifier(self):
        g = Guideline()
        self.assertIsNone(g.identifier)
        g.generateIdentifier()
        self.assertIsNotNone(g.identifier)
        self.assertTrue(g.dirty)

    def test_dirty(self):
        g = Guideline()
        self.assertFalse(g.dirty)

    def test_instance(self):
        g = Guideline(guidelineDict=dict(x=1, y=2, angle=3, name="4",
                                         identifier="5", color="1,1,1,1"))
        self.assertEqual((g.x, g.y, g.angle, g.name, g.identifier, g.color),
                         (1, 2, 3, '4', '5', '1,1,1,1'))


if __name__ == "__main__":
    unittest.main()



from __future__ import absolute_import


class Point(object):

    """
    This object represents a single point.
    """

    __slots__ = ["_x", "_y", "_segmentType", "_smooth", "_name", "_identifier"]

    def __init__(self, coordinates, segmentType=None, smooth=False, name=None, identifier=None):
        (x, y) = coordinates
        super(Point, self).__init__()
        self._x = x
        self._y = y
        self._segmentType = segmentType
        self._smooth = smooth
        self._name = name
        self._identifier = identifier

    def __repr__(self):
        return "<%s position: (%s, %s) type: %s smooth: %s name: %s identifier: %s>" % (self.__class__.__name__, self.x, self.y, str(self.segmentType), str(self.smooth), str(self.name), str(self.identifier))

    def _get_segmentType(self):
        return self._segmentType

    def _set_segmentType(self, value):
        self._segmentType = value

    segmentType = property(_get_segmentType, _set_segmentType, doc="The segment type. The positibilies are *move*, *line*, *curve*, *qcurve* and *None* (indicating that this is an off-curve point).")

    def _get_x(self):
        return self._x

    def _set_x(self, value):
        self._x = value

    x = property(_get_x, _set_x, doc="The x coordinate.")

    def _get_y(self):
        return self._y

    def _set_y(self, value):
        self._y = value

    y = property(_get_y, _set_y, doc="The y coordinate.")

    def _get_smooth(self):
        return self._smooth

    def _set_smooth(self, value):
        self._smooth = value

    smooth = property(_get_smooth, _set_smooth, doc="A boolean indicating the smooth state of the point.")

    def _get_name(self):
        return self._name

    def _set_name(self, value):
        self._name = value

    name = property(_get_name, _set_name, doc="An arbitrary name for the point.")

    def move(self, values):
        """
        Move the component by **(x, y)**.
        """
        (x, y) = values
        self.x += x
        self.y += y

    # ----------
    # Identifier
    # ----------

    def _get_identifier(self):
        return self._identifier

    def _set_identifier(self, value):
        # don't allow overwritting an existing identifier
        if self._identifier is not None:
            return
        self._identifier = value

    identifier = property(_get_identifier, _set_identifier, doc="The identifier.")


if __name__ == "__main__":
    import doctest
    doctest.testmod()



from __future__ import absolute_import
import weakref
import pickle

class BaseObject(object):

    """
    The base object in defcon from which all other objects should be derived.

    **This object posts the following notifications:**

    - BaseObject.Changed
    - BaseObject.BeginUndo
    - BaseObject.EndUndo
    - BaseObject.BeginRedo
    - BaseObject.EndRedo

    Keep in mind that subclasses will not post these same notifications.

    Subclasses must override the following attributes:

    +-------------------------+--------------------------------------------------+
    | Name                    | Notes                                            |
    +=========================+==================================================+
    | changeNotificationName  | This must be a string unique to the class        |
    |                         | indicating the name of the notification          |
    |                         | to be posted when the dirty attribute is set.    |
    +-------------------------+--------------------------------------------------+
    | representationFactories | This must be a dictionary that is shared across  |
    |                         | *all* instances of the class.                    |
    +-------------------------+--------------------------------------------------+
    """

    changeNotificationName = "BaseObject.Changed"
    beginUndoNotificationName = "BaseObject.BeginUndo"
    endUndoNotificationName = "BaseObject.EndUndo"
    beginRedoNotificationName = "BaseObject.BeginRedo"
    endRedoNotificationName = "BaseObject.EndRedo"
    representationFactories = None

    def __init__(self):
        self._init()

    def _init(self):
        self._dispatcher = None
        self._dataOnDisk = None
        self._dataOnDiskTimeStamp = None
        self._undoManager = None
        self._representations = {}

    def __del__(self):
        self.endSelfNotificationObservation()

    # ------
    # Parent
    # ------

    def getParent(self):
        raise NotImplementedError

    # -------------
    # Notifications
    # -------------

    def _get_dispatcher(self):
        if self._dispatcher is not None:
            return self._dispatcher()
        else:
            try:
                dispatcher = self.font.dispatcher
                self._dispatcher = weakref.ref(dispatcher)
            except AttributeError:
                dispatcher = None
        return dispatcher

    dispatcher = property(_get_dispatcher, doc="The :class:`defcon.tools.notifications.NotificationCenter` assigned to the parent of this object.")

    def addObserver(self, observer, methodName, notification, identifier=None):
        """
        Add an observer to this object's notification dispatcher.

        * **observer** An object that can be referenced with weakref.
        * **methodName** A string representing the method to be called
          when the notification is posted.
        * **notification** The notification that the observer should
          be notified of.
        * **identifier** None or a string identifying the observation.
          There is no requirement that the string be unique. A reverse
          domain naming scheme is recommended, but there are no
          requirements for the structure of the string.

        The method that will be called as a result of the action
        must accept a single *notification* argument. This will
        be a :class:`defcon.tools.notifications.Notification` object.

        This is a convenience method that does the same thing as::

            dispatcher = anObject.dispatcher
            dispatcher.addObserver(observer=observer, methodName=methodName,
                notification=notification, observable=anObject, identifier=identifier)
        """
        dispatcher = self.dispatcher
        if dispatcher is not None:
            self.dispatcher.addObserver(observer=observer, methodName=methodName,
                notification=notification, observable=self, identifier=identifier)

    def removeObserver(self, observer, notification):
        """
        Remove an observer from this object's notification dispatcher.

        * **observer** A registered object.
        * **notification** The notification that the observer was registered
          to be notified of.

        This is a convenience method that does the same thing as::

            dispatcher = anObject.dispatcher
            dispatcher.removeObserver(observer=observer,
                notification=notification, observable=anObject)
        """
        dispatcher = self.dispatcher
        if dispatcher is not None:
            self.dispatcher.removeObserver(observer=observer, notification=notification, observable=self)

    def hasObserver(self, observer, notification):
        """
        Returns a boolean indicating is the **observer** is registered for **notification**.

        This is a convenience method that does the same thing as::

            dispatcher = anObject.dispatcher
            dispatcher.hasObserver(observer=observer,
                notification=notification, observable=anObject)
        """
        dispatcher = self.dispatcher
        if dispatcher is not None:
            return self.dispatcher.hasObserver(observer=observer, notification=notification, observable=self)
        return False

    def holdNotifications(self, notification=None, note=None):
        """
        Hold this object's notifications until told to release them.

        * **notification** The specific notification to hold. This is optional.
          If no *notification* is given, all notifications will be held.
        * **note** An arbitrary string containing information about why the hold
          has been requested, the requester, etc. This is used for reference only.

        This is a convenience method that does the same thing as::

            dispatcher = anObject.dispatcher
            dispatcher.holdNotifications(
                observable=anObject, notification=notification, note=note)
        """
        dispatcher = self.dispatcher
        if dispatcher is not None:
            dispatcher.holdNotifications(observable=self, notification=notification, note=note)

    def releaseHeldNotifications(self, notification=None):
        """
        Release this object's held notifications.

        * **notification** The specific notification to hold. This is optional.

        This is a convenience method that does the same thing as::

            dispatcher = anObject.dispatcher
            dispatcher.releaseHeldNotifications(
                observable=anObject, notification=notification)
        """
        dispatcher = self.dispatcher
        if dispatcher is not None:
            dispatcher.releaseHeldNotifications(observable=self, notification=notification)

    def disableNotifications(self, notification=None, observer=None):
        """
        Disable this object's notifications until told to resume them.

        * **notification** The specific notification to disable. This is optional.
          If no *notification* is given, all notifications will be disabled.

        This is a convenience method that does the same thing as::

            dispatcher = anObject.dispatcher
            dispatcher.disableNotifications(
                observable=anObject, notification=notification, observer=observer)
        """
        dispatcher = self.dispatcher
        if dispatcher is not None:
            dispatcher.disableNotifications(observable=self, notification=notification, observer=observer)

    def enableNotifications(self, notification=None, observer=None):
        """
        Enable this object's notifications.

        * **notification** The specific notification to enable. This is optional.

        This is a convenience method that does the same thing as::

            dispatcher = anObject.dispatcher
            dispatcher.enableNotifications(
                observable=anObject, notification=notification, observer=observer)
        """
        dispatcher = self.dispatcher
        if dispatcher is not None:
            dispatcher.enableNotifications(observable=self, notification=notification, observer=observer)

    def postNotification(self, notification, data=None):
        """
        Post a **notification** through this object's notification dispatcher.

            * **notification** The name of the notification.
            * **data** Arbitrary data that will be stored in the :class:`Notification` object.

        This is a convenience method that does the same thing as::

            dispatcher = anObject.dispatcher
            dispatcher.postNotification(
                notification=notification, observable=anObject, data=data)
        """
        dispatcher = self.dispatcher
        if dispatcher is not None:
            dispatcher.postNotification(notification=notification, observable=self, data=data)

    def findObservations(self, observer=None, notification=None, observable=None, identifier=None):
        """
        Find observations of this object matching the given
        arguments based on the values that were passed during
        addObserver. A value of None for any of these indicates
        that all should be considered to match the value.
        In the case of identifier, strings will be matched
        using fnmatch.fnmatchcase. The returned value will be
        a list of dictionaries with this format:

            [
                {
                    observer=<...>
                    observable=<...>
                    methodName="..."
                    notification="..."
                    identifier="..."
                }
            ]

        This is a convenience method that does the same thing as::

            dispatcher = anObject.dispatcher
            dispatcher.findObservations(
                observer=observer, observable=anObject,
                notification=notification, identifier=identifier
            )
        """
        dispatcher = self.dispatcher
        if dispatcher is not None:
            return self.dispatcher.findObservations(observer=observer, notification=notification,
                observable=self, identifier=identifier)

    # ------------------------
    # Notification Observation
    # ------------------------

    def beginSelfNotificationObservation(self):
        self.addObserver(self, "selfNotificationCallback", notification=None)

    def endSelfNotificationObservation(self):
        self.removeObserver(self, notification=None)
        self._dispatcher = None

    def selfNotificationCallback(self, notification):
        self._destroyRepresentationsForNotification(notification)

    # ----
    # Undo
    # ----

    # manager

    def _get_undoManager(self):
        return self._undoManager

    def _set_undoManager(self, manager):
        self._undoManager = manager

    undoManager = property(_get_undoManager, _set_undoManager,
                           doc="The undo manager assigned to this object.")

    # undo

    def canUndo(self):
        """
        Returns a boolean indicating whether the undo manager is able to
        perform an undo.
        """
        return self.undoManager.canUndo()

    def undo(self):
        """
        Perform an undo if possible, or return.
        If undo is performed, this will post *BaseObject.BeginUndo* and *BaseObject.EndUndo* notifications.
        """
        if not self.undoManager.canUndo():
            return
        self.postNotification(notification=self.beginUndoNotificationName)
        self.undoManager.undo()
        self.postNotification(notification=self.endUndoNotificationName)

    # redo

    def canRedo(self):
        """
        Returns a boolean indicating whether the undo manager is able to
        perform a redo.
        """
        return self.undoManager.canRedo()

    def redo(self):
        """
        Perform a redo if possible, or return.
        If redo is performed, this will post *BaseObject.BeginRedo* and *BaseObject.EndRedo* notifications.
        """
        if not self.undoManager.canRedo():
            return
        self.postNotification(notification=self.beginRedoNotificationName)
        self.undoManager.redo()
        self.postNotification(notification=self.endRedoNotificationName)

    # ---------------
    # Representations
    # ---------------

    def getRepresentation(self, name, **kwargs):
        """
        Get a representation. **name** must be a registered
        representation name. **\*\*kwargs** will be passed
        to the appropriate representation factory.
        """
        # only store the representation if the object has a
        # dispatcher. otherwise, notifications may not be
        # destroyed after an object change.
        if self.dispatcher is None:
            factory = self.representationFactories[name]
            return factory["factory"](self, **kwargs)
        else:
            if name not in self._representations:
                self._representations[name] = {}
            representations = self._representations[name]
            subKey = self._makeRepresentationSubKey(**kwargs)
            if subKey not in representations:
                factory = self.representationFactories[name]
                representation = factory["factory"](self, **kwargs)
                representations[subKey] = representation
            return representations[subKey]

    def destroyRepresentation(self, name, **kwargs):
        """
        Destroy the stored representation for **name**
        and **\*\*kwargs**. If no **kwargs** are given,
        any representation with **name** will be destroyed
        regardless of the **kwargs** passed when the
        representation was created.
        """
        if name not in self._representations:
            return
        if not kwargs:
            del self._representations[name]
        else:
            representations = self._representations[name]
            subKey = self._makeRepresentationSubKey(**kwargs)
            if subKey in representations:
                del self._representations[name][subKey]

    def destroyAllRepresentations(self, notification=None):
        """
        Destroy all representations.
        """
        self._representations.clear()

    def _destroyRepresentationsForNotification(self, notification):
        notificationName = notification.name
        for name, dataDict in self.representationFactories.items():
            if notificationName in dataDict["destructiveNotifications"]:
                self.destroyRepresentation(name)

    def representationKeys(self):
        """
        Get a list of all representation keys that are
        currently cached.
        """
        representations = []
        for name, subDict in self._representations.items():
            for subKey in subDict.keys():
                kwargs = {}
                if subKey is not None:
                    for k, v in subKey:
                        kwargs[k] = v
                representations.append((name, kwargs))
        return representations

    def hasCachedRepresentation(self, name, **kwargs):
        """
        Returns a boolean indicating if a representation for
        **name** and **\*\*kwargs** is cached in the object.
        """
        if name not in self._representations:
            return False
        subKey = self._makeRepresentationSubKey(**kwargs)
        return subKey in self._representations[name]

    def _makeRepresentationSubKey(self, **kwargs):
        if kwargs:
            key = sorted(kwargs.items())
            key = tuple(key)
        else:
            key = None
        return key

    # -----
    # Dirty
    # -----

    def _set_dirty(self, value):
        self._dirty = value
        dispatcher = self.dispatcher
        if dispatcher is not None:
            self.postNotification(self.changeNotificationName)

    def _get_dirty(self):
        return self._dirty

    dirty = property(_get_dirty, _set_dirty, doc="The dirty state of the object. True if the object has been changed. False if not. Setting this to True will cause the base changed notification to be posted. The object will automatically maintain this attribute and update it as you change the object.")

    # -----------------------------
    # Serialization/Deserialization
    # -----------------------------

    def serialize(self, dumpFunc=None, whitelist=None, blacklist=None):
        data = self.getDataForSerialization(whitelist=whitelist, blacklist=blacklist)

        dump = dumpFunc if dumpFunc is not None else pickle.dumps
        return dump(data)

    def deserialize(self, data, loadFunc=None):
        load = loadFunc if loadFunc is not None else pickle.loads
        self.setDataFromSerialization(load(data))

    def getDataForSerialization(self, **kwargs):
        """
        Return a dict of data that can be pickled.
        """
        return {}

    def setDataFromSerialization(self, data):
        """
        Restore state from the provided data-dict.
        """
        pass

    def _serialize(self, getters, whitelist=None, blacklist=None, **kwargs):
        """ A helper function for the defcon objects.

        Return a dict where the keys are the keys in getters and the values
        are the results of the getter functions

        getters is a list of tuples:
        [
            (:str:key, :callable:getter_function)
        ]

        if a whitelist is not None: the key must be in whitelist
        if a blacklist is not None: the key must not be in blacklist
        """
        data = {}
        for key, getter in getters:
            if whitelist is not None and key not in whitelist:
                continue
            if blacklist is not None and key in blacklist:
                continue
            data[key] = getter(key)
        return data

    # =============================================
    # = ufo lib writer/reader validation settings =
    # =============================================

    ufoLibReadValidate = True
    ufoLibWriteValidate = True


class BaseDictObject(dict, BaseObject):

    """
    A subclass of BaseObject that implements a dict API. Any changes
    to the contents of the object will cause the dirty attribute
    to be set to True.
    """

    setItemNotificationName = None
    deleteItemNotificationName = None
    clearNotificationName = None
    updateNotificationName = None

    def __init__(self):
        super(BaseDictObject, self).__init__()
        self._init()
        self._dirty = False

    def _get_dict(self):
        from warnings import warn
        warn(
            "BaseDictObject is now a dict and _dict is gone.",
            DeprecationWarning
        )
        return self

    _dict = property(_get_dict)

    def __hash__(self):
        return id(self)

    def __setitem__(self, key, value):
        oldValue = None
        if key in self:
            oldValue = self[key]
            if value is not None and oldValue == value:
                # don't do this if the value is None since some
                # subclasses establish their keys at startup with
                # self[key] = None
                return
        super(BaseDictObject, self).__setitem__(key, value)
        if self.setItemNotificationName is not None:
            self.postNotification(self.setItemNotificationName, data=dict(key=key, oldValue=oldValue, newValue=value))
        self.dirty = True

    def __delitem__(self, key):
        super(BaseDictObject, self).__delitem__(key)
        if self.deleteItemNotificationName is not None:
            self.postNotification(self.deleteItemNotificationName, data=dict(key=key))
        self.dirty = True

    def __deepcopy__(self, memo={}):
        import copy
        obj = self.__class__()
        for k, v in self.items():
            k = copy.deepcopy(k)
            v = copy.deepcopy(v)
            obj[k] = v
        return obj

    def clear(self):
        if not len(self):
            return
        super(BaseDictObject, self).clear()
        if self.clearNotificationName is not None:
            self.postNotification(self.clearNotificationName)
        self.dirty = True

    def update(self, other):
        super(BaseDictObject, self).update(other)
        if self.updateNotificationName is not None:
            self.postNotification(self.updateNotificationName, data=dict(other=other))
        self.dirty = True

    # -----------------------------
    # Serialization/Deserialization
    # -----------------------------

    def getDataForSerialization(self, **kwargs):
        simple_get = lambda k: self[k]

        getters = []
        for k in self.keys():
            getters.append((k, simple_get))

        return self._serialize(getters, **kwargs)

    def setDataFromSerialization(self, data):
        self.clear()
        self.update(data)


class BaseDictCompareObject(BaseDictObject):

    # hash and eq must happen on the same class
    # otherwise the object becomes unhashable
    def __hash__(self):
        return id(self)

    def __eq__(self, other):
        return id(self) == id(other)


def setUfoLibReadValidate(value):
    """
    Set the default read validation.
    """
    BaseObject.ufoLibReadValidate = value


def setUfoLibWriteValidate(value):
    """
    Set the default write validation.
    """
    BaseObject.ufoLibWriteValidate = value


if __name__ == "__main__":
    import doctest
    doctest.testmod()



from __future__ import absolute_import
import os
import hashlib
import weakref
from defcon.objects.base import BaseObject
from fontTools.ufoLib import UFOReader, UFOLibError
from fontTools.ufoLib.filenames import (
    userNameToFileName, illegalCharacters, reservedFileNames, maxFileNameLength
)
from fontTools.ufoLib.validators import pngSignature


class ImageSet(BaseObject):

    """
    This object manages all images in the font.

    **This object posts the following notifications:**

    - ImageSet.Changed
    - ImageSet.FileNamesChanged
    - ImageSet.ImageChanged
    - ImageSet.ImageWillBeAdded
    - ImageSet.ImageAdded
    - ImageSet.ImageWillBeDeleted
    - ImageSet.ImageDeleted

    This object behaves like a dict. For example, to get the
    raw image data for a particular image::

        image = images["image file name"]

    To add an image, do this::

        images["image file name"] = rawImageData

    When setting an image, the provided file name must be a file
    system legal string. This will be checked by comparing the
    provided file name to the results of :py:meth:`ImageSet.makeFileName`.
    If the two don't match an error will be raised.

    Before setting an image, the :py:meth:`ImageSet.findDuplicateImage`
    method should be called. If a file name is retruend, the new image
    data should not be added. The UFO spec recommends (but doesn't require)
    that duplicate images be avoided. This will help with that.

    To remove an image from this object, and from the UFO during save,
    do this::

        del images["image file name"]
    """

    changeNotificationName = "ImageSet.Changed"
    representationFactories = {}

    def __init__(self, font=None):
        self._font = None
        if font is not None:
            self._font = weakref.ref(font)
        super(ImageSet, self).__init__()
        self.beginSelfNotificationObservation()
        self._dirty = False
        self._data = {}
        self._scheduledForDeletion = {}

    # --------------
    # Parent Objects
    # --------------

    def getParent(self):
        return self.font

    def _get_font(self):
        if self._font is not None:
            return self._font()
        return None

    font = property(_get_font, doc="The :class:`Font` that this object belongs to.")

    # ----------
    # File Names
    # ----------

    def _get_fileNames(self):
        return list(self._data.keys())

    def _set_fileNames(self, fileNames):
        assert not self._data
        oldValue = list(self._data.keys())
        for fileName in fileNames:
            self._data[fileName] = _imageDict(onDisk=True)
        self.postNotification("ImageSet.FileNamesChanged", data=dict(oldValue=oldValue, newValue=fileNames))

    fileNames = property(_get_fileNames, _set_fileNames, doc="A list of all image file names. This should not be set externally.")

    def _get_unreferencedFileNames(self):
        font = self.font
        if font is None:
            return []
        unreferenced = set(self.fileNames)
        for layer in font.layers:
            unreferenced -= set(layer.imageReferences.keys())
        return list(unreferenced)

    unreferencedFileNames = property(_get_unreferencedFileNames, doc="A list of all file names not referenced by a glyph.")

    # -------------
    # Dict Behavior
    # -------------

    def __contains__(self, fileName):
        return fileName in self._data

    def __getitem__(self, fileName):
        d = self._data[fileName]
        if d["data"] is None:
            reader = self.font._reader
            data = reader.readImage(fileName, validate=self.ufoLibReadValidate)
            d["data"] = data
            d["digest"] = _makeDigest(data)
            d["onDisk"] = True
            d["onDiskModTime"] = reader.getFileModificationTime("%s/%s" % ("images", fileName))
        return d["data"]

    def __setitem__(self, fileName, data):
        if fileName not in self._data:
            test = fileName
            if fileName.lower().endswith(".png"):
                test = os.path.splitext(fileName)[0]
            assert fileNameValidator(test)
        assert data.startswith(pngSignature), "Image does not begin with the PNG signature."
        isNewImage = fileName not in self._data
        onDisk = False
        onDiskModTime = None
        if fileName in self._scheduledForDeletion:
            # preserve exsiting stamping
            assert fileName not in self._data
            self._data[fileName] = self._scheduledForDeletion.pop(fileName)
        digest = _makeDigest(data)
        if fileName in self._data:
            n = self[fileName] # force it to load so that the stamping is correct
            if self._data[fileName]["digest"] == digest:
                return
            onDisk = self._data[fileName]["onDisk"]
            onDiskModTime = self._data[fileName]["onDiskModTime"]
            del self._data[fileName] # now remove it
        if isNewImage:
            self.postNotification("ImageSet.ImageWillBeAdded", data=dict(name=fileName))
        self._data[fileName] = _imageDict(data=data, dirty=True, digest=digest, onDisk=onDisk, onDiskModTime=onDiskModTime)
        if isNewImage:
            self.postNotification("ImageSet.ImageAdded", data=dict(name=fileName))
        else:
            self.postNotification("ImageSet.ImageChanged", data=dict(name=fileName))
        self.dirty = True

    def __delitem__(self, fileName):
        n = self[fileName] # force it to load so that the stamping is correct
        self.postNotification("ImageSet.ImageWillBeDeleted", data=dict(name=fileName))
        self._scheduledForDeletion[fileName] = dict(self._data.pop(fileName))
        self.postNotification("ImageSet.ImageDeleted", data=dict(name=fileName))
        self.dirty = True

    # ----
    # Save
    # ----

    def getSaveProgressBarTickCount(self, formatVersion):
        """
        Get the number of ticks that will be used by a progress bar
        in the save method. This method should not be called externally.
        Subclasses may override this method to implement custom saving behavior.
        """
        return 0

    def save(self, writer, removeUnreferencedImages=False, saveAs=False, progressBar=None):
        """
        Save images. This method should not be called externally.
        Subclasses may override this method to implement custom saving behavior.
        """
        if removeUnreferencedImages:
            self.disableNotifications()
            for fileName in self.unreferencedFileNames:
                del self[fileName]
            self.enableNotifications()
        if saveAs:
            font = self.font
            if font is not None and font.path is not None and os.path.exists(font.path):
                with UFOReader(font.path, validate=False) as reader:
                    readerImageNames = reader.getImageDirectoryListing(validate=self.ufoLibReadValidate)
                    for fileName, data in self._data.items():
                        if data["data"] is not None or fileName not in readerImageNames:
                            continue
                        writer.copyImageFromReader(reader, fileName, fileName, validate=self.ufoLibWriteValidate)
        for fileName in self._scheduledForDeletion:
            try:
                writer.removeImage(fileName, validate=self.ufoLibWriteValidate)
            except UFOLibError:
                # this will be raised if the file doesn't exist.
                # instead of trying to maintain a list of in UFO
                # vs. in memory, simply fail and move on when
                # something can't be deleted because it isn't
                # in the UFO.
                pass
        self._scheduledForDeletion.clear()
        for fileName, data in self._data.items():
            if not data["dirty"]:
                continue
            writer.writeImage(fileName, data["data"], validate=self.ufoLibWriteValidate)
            data["dirty"] = False
            data["onDisk"] = True
            data["onDiskModTime"] = writer.getFileModificationTime("%s/%s" % ("images", fileName))
        self.dirty = False

    # ---------------
    # File Management
    # ---------------

    def makeFileName(self, fileName):
        """
        Make a file system legal version of **fileName**.
        """
        fileName = str(fileName)
        suffix = ""
        if fileName.lower().endswith(".png"):
            suffix = fileName[-4:]
            fileName = fileName[:-4]
        existing = set([i.lower() for i in self.fileNames])
        return userNameToFileName(fileName, existing, suffix=suffix)

    def findDuplicateImage(self, data):
        """
        Search the images to see if an image matching
        **image** already exists. If so, the file name
        for the existing image will be returned.
        """
        digest = _makeDigest(data)
        notYetLoaded = []
        for fileName, image in self._data.items():
            # skip if the image hasn't been loaded
            if image["data"] is None:
                notYetLoaded.append(fileName)
                continue
            otherDigest = image["digest"]
            if otherDigest == digest:
                return fileName
        for fileName in notYetLoaded:
            d = self[fileName]
            image = self._data[fileName]
            otherDigest = image["digest"]
            if otherDigest == digest:
                return fileName
        return None

    # ---------------------
    # External Edit Support
    # ---------------------

    def testForExternalChanges(self, reader):
        """
        Test for external changes. This should not be called externally.
        """
        filesOnDisk = reader.getImageDirectoryListing()
        modifiedImages = []
        addedImages = []
        deletedImages = []
        for fileName in set(filesOnDisk) - set(self.fileNames):
            if fileName not in self._scheduledForDeletion:
                addedImages.append(fileName)
            elif not self._scheduledForDeletion[fileName]["onDisk"]:
                addedImages.append(fileName)
            elif self._scheduledForDeletion[fileName]["onDiskModTime"] != reader.getFileModificationTime(
                "%s/%s" % ("images", fileName)
            ):
                addedImages.append(fileName)
        for fileName, imageData in self._data.items():
            # file on disk and has been loaded
            if fileName in filesOnDisk and imageData["data"] is not None:
                newModTime = reader.getFileModificationTime("%s/%s" % ("images", fileName))
                if newModTime != imageData["onDiskModTime"]:
                    newData = reader.readImage(fileName)
                    newDigest = _makeDigest(newData)
                    if newDigest != imageData["digest"]:
                        modifiedImages.append(fileName)
                continue
            # file removed
            if fileName not in filesOnDisk and imageData["onDisk"]:
                deletedImages.append(fileName)
                continue
        return modifiedImages, addedImages, deletedImages

    def reloadImages(self, fileNames):
        """
        Reload specified images. This should not be called externally.
        """
        for fileName in fileNames:
            self._data[fileName] = _imageDict()
            image = self[fileName]

    # ------------------------
    # Notification Observation
    # ------------------------

    def endSelfNotificationObservation(self):
        super(ImageSet, self).endSelfNotificationObservation()
        self._font = None

    # -----------------------------
    # Serialization/Deserialization
    # -----------------------------

    def getDataForSerialization(self, **kwargs):
        simple_get = lambda k: self[k]

        getters = []
        for k in self.fileNames:
            getters.append((k, simple_get))

        return self._serialize(getters, **kwargs)

    def setDataFromSerialization(self, data):
        self._data = {}
        self._scheduledForDeletion = {}
        for k in data:
            self[k] = data[k]

def _imageDict(data=None, dirty=False, digest=None, onDisk=True, onDiskModTime=None):
    return dict(data=data, digest=digest, dirty=dirty, onDisk=onDisk, onDiskModTime=onDiskModTime)

def _makeDigest(data):
    m = hashlib.md5()
    m.update(data)
    return m.digest()

def fileNameValidator(value):
    """
    >>> fileNameValidator(u'a')
    True
    >>> fileNameValidator(u'A_')
    True
    >>> fileNameValidator(u'A_E_')
    True
    >>> fileNameValidator(u'A_e')
    True
    >>> fileNameValidator(u'ae')
    True
    >>> fileNameValidator(u'aE_')
    True
    >>> fileNameValidator(u'a.alt')
    True
    >>> fileNameValidator(u'A_.alt')
    True
    >>> fileNameValidator(u'A_.A_lt')
    True
    >>> fileNameValidator(u'A_.aL_t')
    True
    >>> fileNameValidator(u'A_.alT_')
    True
    >>> fileNameValidator(u'T__H_')
    True
    >>> fileNameValidator(u'T__h')
    True
    >>> fileNameValidator(u't_h')
    True
    >>> fileNameValidator(u'F__F__I_')
    True
    >>> fileNameValidator(u'f_f_i')
    True
    >>> fileNameValidator(u'A_acute_V_.swash')
    True
    >>> fileNameValidator(u'_notdef')
    True
    >>> fileNameValidator(u'_con')
    True
    >>> fileNameValidator(u'C_O_N_')
    True
    >>> fileNameValidator(u'_con.alt')
    True
    >>> fileNameValidator(u'alt._con')
    True
    >>> fileNameValidator('A')
    False
    >>> fileNameValidator(u'A'*256)
    False
    >>> fileNameValidator(u'A')
    False
    >>> fileNameValidator(u'con')
    False
    >>> fileNameValidator(u'a/alt')
    False
    >>> fileNameValidator(u"A_bC_dE_f")
    True
    """
    # must be a unicode
    if not isinstance(value, str):
        return False
    # must not be longer then the max fileName length
    if len(value) > maxFileNameLength:
        return False
    for i, character in enumerate(value):
        # must not contain any illegal characters
        if character in illegalCharacters:
            return False
        # its a capital and it should be followed by an _ (underscore)
        elif character != character.lower():
            if i == len(value)-1:
                return False
            if value[i+1] != "_":
                return False
    # check reserved file names
    for reservedFileName in reservedFileNames:
        # all reserved file names are being prefix with an _ (underscore)
        # if the replaced value is the same there is no correct prefix
        if reservedFileName in value:
            if value == value.replace("_%s" % reservedFileName, ""):
                return False
    return True

if __name__ == "__main__":
    import doctest
    doctest.testmod()



from __future__ import absolute_import
import weakref
from defcon.objects.base import BaseDictObject
from defcon.objects.color import Color


_defaultImage = {
    "fileName": None,
    "color"   : None,
    "xScale"  : 1,
    "xyScale" : 0,
    "yxScale" : 0,
    "yScale"  : 1,
    "xOffset" : 0,
    "yOffset" : 0
}

class Image(BaseDictObject):

    """
    This object represents an image reference in a glyph.

    **This object posts the following notifications:**

    - Image.Changed
    - Image.FileNameChanged
    - Image.TransformationChanged
    - Image.ColorChanged
    - Image.ImageDataChanged

    During initialization an image dictionary, following the format defined
    in the UFO spec, can be passed. If so, the new object will be populated
    with the data from the dictionary.
    """

    changeNotificationName = "Image.Changed"
    representationFactories = {}

    def __init__(self, glyph=None, imageDict=None):
        self._font = None
        self._layerSet = None
        self._layer = None
        self._glyph = None
        self.glyph = glyph
        super(Image, self).__init__()
        self.beginSelfNotificationObservation()
        for key, value in _defaultImage.items():
            if self.get(key) is None:
                self[key] = value
        if imageDict is not None:
            self.update(imageDict)
        self._dirty = False

    def __len__(self):
        # this is a little hack for glifLib writing.
        # when a GLIF is written, glyph.image is chekced with:
        #     if glyph.image:
        # fileName is required, so if that isn't defined
        # return 0. this tells glifLib to skip the image.
        if self["fileName"] is None:
            return 0
        return super(Image, self).__len__()

    def clear(self):
        # clear all the image data and fill it with default values
        super(BaseDictObject, self).clear()
        for key, value in _defaultImage.items():
            if self.get(key) is None:
                self[key] = value
    # --------------
    # Parent Objects
    # --------------

    def getParent(self):
        return self.glyph

    def _get_font(self):
        font = None
        if self._font is None:
            glyph = self.glyph
            if glyph is not None:
                font = glyph.font
                if font is not None:
                    self._font = weakref.ref(font)
        else:
            font = self._font()
        return font

    font = property(_get_font, doc="The :class:`Font` that this image belongs to.")

    def _get_layerSet(self):
        layerSet = None
        if self._layerSet is None:
            glyph = self.glyph
            if glyph is not None:
                layerSet = glyph.layerSet
                if layerSet is not None:
                    self._layerSet = weakref.ref(layerSet)
        else:
            layerSet = self._layerSet()
        return layerSet

    layerSet = property(_get_layerSet, doc="The :class:`LayerSet` that this image belongs to.")

    def _get_layer(self):
        layer = None
        if self._layer is None:
            glyph = self.glyph
            if glyph is not None:
                layer = glyph.layer
                if layer is not None:
                    self._layer = weakref.ref(layer)
        else:
            layer = self._layer()
        return layer

    layer = property(_get_layer, doc="The :class:`Layer` that this image belongs to.")

    def _get_glyph(self):
        if self._glyph is None:
            return None
        return self._glyph()

    def _set_glyph(self, glyph):
        assert self._glyph is None
        if glyph is not None:
            glyph = weakref.ref(glyph)
        self._font = None
        self._layerSet = None
        self._layer = None
        self._glyph = glyph

    glyph = property(_get_glyph, _set_glyph, doc="The :class:`Glyph` that this image belongs to. This should not be set externally.")

    # ----------
    # Attributes
    # ----------

    # file name

    def _get_fileName(self):
        return self["fileName"]

    def _set_fileName(self, fileName):
        oldFileName = self.get("fileName")
        if fileName == oldFileName:
            return
        self["fileName"] = fileName
        self.postNotification("Image.FileNameChanged", data=dict(oldValue=oldFileName, newValue=fileName))

    fileName = property(_get_fileName, _set_fileName, doc="The file name the image. Setting this will posts *Image.Changed* and *Image.FileNameChanged* notifications.")

    # transformation

    def _get_transformation(self):
        if "xScale" not in self:
            return
        return (self["xScale"], self["xyScale"], self["yxScale"], self["yScale"], self["xOffset"], self["yOffset"])

    def _set_transformation(self, transformation):
        oldTransformation = self.transformation
        if oldTransformation == transformation:
            return
        xScale, xyScale, yxScale, yScale, xOffset, yOffset = transformation
        # hold the notifications so that only one is sent out
        self.holdNotifications(note="Requested by Image._set_transformation.")
        self["xScale"] = xScale
        self["xyScale"] = xyScale
        self["yxScale"] = yxScale
        self["yScale"] = yScale
        self["xOffset"] = xOffset
        self["yOffset"] = yOffset
        self.releaseHeldNotifications()
        self.postNotification("Image.TransformationChanged", data=dict(oldValue=oldTransformation, newValue=transformation))

    transformation = property(_get_transformation, _set_transformation, doc="The transformation matrix for the image. Setting this will posts *Image.Changed* and *Image.TransformationChanged* notifications.")

    # color

    def _get_color(self):
        return self.get("color")

    def _set_color(self, color):
        if color is None:
            newColor = None
        else:
            newColor = Color(color)
        oldColor = self.get("color")
        if newColor == oldColor:
            return
        self["color"] = newColor
        self.postNotification("Image.ColorChanged", data=dict(oldValue=oldColor, newValue=newColor))

    color = property(_get_color, _set_color, doc="The image's :class:`Color` object. When setting, the value can be a UFO color string, a sequence of (r, g, b, a) or a :class:`Color` object. Setting this posts *Image.ColorChanged* and *Image.Changed* notifications.")

    # ----
    # Move
    # ----

    def move(self, values):
        """
        Move the image by **(x, y)**.

        This posts *Image.Changed* and *Image.TransformationChanged* notifications.
        """
        xOffset, yOffset = values
        if not (xOffset or yOffset):
            return
        oldTransformation = self.transformation
        self.holdNotifications(note="Requested by Image.move.")
        self["xOffset"] += xOffset
        self["yOffset"] += yOffset
        self.releaseHeldNotifications()
        self.postNotification("Image.TransformationChanged", data=dict(oldValue=oldTransformation, newValue=self.transformation))

    # ------------------------
    # Notification Observation
    # ------------------------

    def beginSelfNotificationObservation(self):
        super(Image, self).beginSelfNotificationObservation()
        self.beginSelfImageSetNotificationObservation()

    def endSelfNotificationObservation(self):
        self.endImageSetNotificationObservation()
        super(Image, self).endSelfNotificationObservation()
        self._font = None
        self._layerSet = None
        self._layer = None
        self._glyph = None

    def beginSelfImageSetNotificationObservation(self):
        font = self.font
        if font is None:
            return
        imageSet = font.images
        imageSet.addObserver(self, "imageSetImageAddedNotificationCallback", "ImageSet.ImageAdded")
        imageSet.addObserver(self, "imageSetImageDeletedNotificationCallback", "ImageSet.ImageDeleted")
        imageSet.addObserver(self, "imageSetImageChangedNotificationCallback", "ImageSet.ImageChanged")
        layer = self.layer
        layer.addObserver(self, "layerColorChangedNotificationCallback", "Layer.ColorChanged")

    def endImageSetNotificationObservation(self):
        font = self.font
        if font is None:
            return
        imageSet = font.images
        imageSet.removeObserver(self, "ImageSet.ImageAdded")
        imageSet.removeObserver(self, "ImageSet.ImageDeleted")
        imageSet.removeObserver(self, "ImageSet.ImageChanged")
        layer = self.layer
        layer.removeObserver(self, "Layer.ColorChanged")

    def imageSetImageAddedNotificationCallback(self, notification):
        name = notification.data["name"]
        if name != self.fileName:
            return
        self.postNotification("Image.ImageDataChanged")

    def imageSetImageDeletedNotificationCallback(self, notification):
        name = notification.data["name"]
        if name != self.fileName:
            return
        self.postNotification("Image.ImageDataChanged")

    def imageSetImageChangedNotificationCallback(self, notification):
        name = notification.data["name"]
        if name != self.fileName:
            return
        self.postNotification("Image.ImageDataChanged")

    def layerColorChangedNotificationCallback(self, notification):
        if self.color is not None:
            self.postNotification("Image.ColorChanged", data=notification.data)


if __name__ == "__main__":
    import doctest
    doctest.testmod()



from __future__ import absolute_import
import weakref
from defcon.objects.base import BaseDictCompareObject
from defcon.objects.color import Color
from defcon.tools.identifiers import makeRandomIdentifier


class Anchor(BaseDictCompareObject):

    """
    This object represents an anchor point.

    **This object posts the following notifications:**

    - Anchor.Changed
    - Anchor.XChanged
    - Anchor.YChanged
    - Anchor.NameChanged
    - Anchor.ColorChanged
    - Anchor.IdentifierChanged

    During initialization an anchor dictionary can be passed. If so,
    the new object will be populated with the data from the dictionary.
    """

    changeNotificationName = "Anchor.Changed"
    representationFactories = {}

    def __init__(self, glyph=None, anchorDict=None):
        self._font = None
        self._layerSet = None
        self._layer = None
        self._glyph = None
        self.glyph = glyph
        super(Anchor, self).__init__()
        self.beginSelfNotificationObservation()
        self._dirty = False
        if anchorDict is not None:
            self.x = anchorDict.get("x")
            self.y = anchorDict.get("y")
            self.name = anchorDict.get("name")
            self.color = anchorDict.get("color")
            self.identifier = anchorDict.get("identifier")

    # parents

    def getParent(self):
        return self.glyph

    def _get_font(self):
        font = None
        if self._font is None:
            glyph = self.glyph
            if glyph is not None:
                font = glyph.font
                if font is not None:
                    self._font = weakref.ref(font)
        else:
            font = self._font()
        return font

    font = property(_get_font, doc="The :class:`Font` that this anchor belongs to.")

    def _get_layerSet(self):
        layerSet = None
        if self._layerSet is None:
            glyph = self.glyph
            if glyph is not None:
                layerSet = glyph.layerSet
                if layerSet is not None:
                    self._layerSet = weakref.ref(layerSet)
        else:
            layerSet = self._layerSet()
        return layerSet

    layerSet = property(_get_layerSet, doc="The :class:`LayerSet` that this anchor belongs to.")

    def _get_layer(self):
        layer = None
        if self._layer is None:
            glyph = self.glyph
            if glyph is not None:
                layer = glyph.layer
                if layer is not None:
                    self._layer = weakref.ref(layer)
        else:
            layer = self._layer()
        return layer

    layer = property(_get_layer, doc="The :class:`Layer` that this anchor belongs to.")

    def _get_glyph(self):
        if self._glyph is None:
            return None
        return self._glyph()

    def _set_glyph(self, glyph):
        assert self._glyph is None
        if glyph is not None:
            glyph = weakref.ref(glyph)
        self._font = None
        self._layerSet = None
        self._layer = None
        self._glyph = glyph


    glyph = property(_get_glyph, _set_glyph, doc="The :class:`Glyph` that this anchor belongs to. This should not be set externally.")

    # coordinates

    def _get_x(self):
        return self.get("x")

    def _set_x(self, value):
        old = self.get("x")
        if value == old:
            return
        self["x"] = value
        self.postNotification("Anchor.XChanged", data=dict(oldValue=old, newValue=value))

    x = property(_get_x, _set_x, doc="The x coordinate. Setting this will post *Anchor.XChanged* and *Anchor.Changed* notifications.")

    def _get_y(self):
        return self.get("y")

    def _set_y(self, value):
        old = self.get("y")
        if value == old:
            return
        self["y"] = value
        self.postNotification("Anchor.YChanged", data=dict(oldValue=old, newValue=value))

    y = property(_get_y, _set_y, doc="The y coordinate. Setting this will post *Anchor.YChanged* and *Anchor.Changed* notifications.")

    # name

    def _get_name(self):
        return self.get("name")

    def _set_name(self, value):
        old = self.get("name")
        if value == old:
            return
        self["name"] = value
        self.postNotification("Anchor.NameChanged", data=dict(oldValue=old, newValue=value))

    name = property(_get_name, _set_name, doc="The name. Setting this will post *Anchor.NameChanged* and *Anchor.Changed* notifications.")

    # color

    def _get_color(self):
        return self.get("color")

    def _set_color(self, color):
        if color is None:
            newColor = None
        else:
            newColor = Color(color)
        oldColor = self.get("color")
        if newColor == oldColor:
            return
        self["color"] = newColor
        self.postNotification("Anchor.ColorChanged", data=dict(oldValue=oldColor, newValue=newColor))

    color = property(_get_color, _set_color, doc="The anchors's :class:`Color` object. When setting, the value can be a UFO color string, a sequence of (r, g, b, a) or a :class:`Color` object. Setting this posts *Anchor.ColorChanged* and *Anchor.Changed* notifications.")

    # identifier

    def _get_identifiers(self):
        identifiers = None
        glyph = self.glyph
        if glyph is not None:
            identifiers = glyph.identifiers
        if identifiers is None:
            identifiers = set()
        return identifiers

    identifiers = property(_get_identifiers, doc="Set of identifiers for the glyph that this anchor belongs to. This is primarily for internal use.")

    def _get_identifier(self):
        return self.get("identifier")

    def _set_identifier(self, value):
        # don't allow overwritting an existing identifier
        if self.identifier is not None:
            return
        oldIdentifier = self.identifier
        if value == oldIdentifier:
            return
        # don't allow a duplicate
        identifiers = self.identifiers
        assert value not in identifiers
        # free the old identifier
        if oldIdentifier in identifiers:
            identifiers.remove(oldIdentifier)
        # store
        self["identifier"] = value
        if value is not None:
            identifiers.add(value)
        # post notifications
        self.postNotification("Anchor.IdentifierChanged", data=dict(oldValue=oldIdentifier, newValue=value))

    identifier = property(_get_identifier, _set_identifier, doc="The identifier. Setting this will post *Anchor.IdentifierChanged* and *Anchor.Changed* notifications.")

    def generateIdentifier(self):
        """
        Create a new, unique identifier for and assign it to the guideline.
        This will post *Anchor.IdentifierChanged* and *Anchor.Changed* notifications.
        """
        if self.identifier is None:
            identifier = makeRandomIdentifier(existing=self.identifiers)
            self.identifier = identifier
        return self.identifier

    # ----
    # Move
    # ----

    def move(self, values):
        """
        Move the anchor by **(x, y)**.

        This will post *Anchor.XChange*, *Anchor.YChanged* and *Anchor.Changed* notifications if anything changed.
        """
        (x, y) = values
        self.x += x
        self.y += y

    # ------------------------
    # Notification Observation
    # ------------------------

    def endSelfNotificationObservation(self):
        super(Anchor, self).endSelfNotificationObservation()
        self._font = None
        self._layerSet = None
        self._layer = None
        self._glyph = None


if __name__ == "__main__":
    import doctest
    doctest.testmod()



"""
Contributed by Frederik Berlaen.
"""

from math import sqrt, atan2, radians, degrees

def _distance(coordinates1, coordinates2):
    (x1, y1) = coordinates1
    (x2, y2) = coordinates2
    return sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2)

def joinSegments(onCoords1, offCoords1, offCoords2, onCoords2, offCoords3, offCoords4, onCoords3):
    """
    >>> joinSegments(
    ...    (0, 0),
    ...    (0, 138), (112, 250), (250, 250),
    ...    (250, 388), (500, 138), (500, 0)
    ...    )
    ((0.0, 276.0), (500.0, 276.0), (500, 0))
    """
    (on1X, on1Y) = onCoords1
    (off1X, off1Y) = offCoords1
    (off2X, off2Y) = offCoords2
    (on2X, on2Y) = onCoords2
    (off3X, off3Y) = offCoords3
    (off4X, off4Y) = offCoords4
    (on3X, on3Y) = onCoords3
    if (on1X, on1Y) == (off1X, off1Y) and (off2X, off2Y) == (on2X, on2Y) == (off3X, off3Y) and (off4X, off4Y) == (on3X, on3Y):
        ## a two line segments
        return (on1X, on1Y), (off4X, off4Y), (on3X, on3Y)
    if (on1X, on1Y) == (off1X, off1Y) and (off2X, off2Y) == (on2X, on2Y):
        ## first is a line segement
        lineAngle = atan2(on2X - on1X, on2Y - on1Y)
        handleAngle = atan2(off3X - on2X, off3Y - on2Y)
        if radians(-30) < lineAngle - handleAngle < radians(30):
            d1 = _distance((on1X, on1Y), (off2X, off2Y))
            d2 = d1 + _distance((on2X, on2Y), (off3X, off3Y))
            if d1 == 0:
                x, y = off3X, off3Y
            else:
                factor = d2 / d1
                x = on1X + (off2X - on1X) * factor
                y = on1Y + (off2Y - on1Y) * factor
            return (x, y), (off4X, off4Y), (on3X, on3Y)
        else:
            # just move the off curve point
            x = off3X - (off2X - on1X)
            y = off3Y - (off2Y - on1Y)
            return (x, y), (off4X, off4Y), (on3X, on3Y)

    if (on2X, on2Y) == (off3X, off3Y) and (off4X, off4Y) == (on3X, on3Y):
        ## last is a line segment
        handleAngle = atan2(on2X - off2X, on2Y - off2Y)
        lineAngle = atan2(on3X - on2X, on3Y - on2Y)
        if radians(-30) < lineAngle - handleAngle < radians(30):
            d1 = _distance((on3X, on3Y), (off3X, off3Y))
            d2 = d1 + _distance((on2X, on2Y), (off2X, off2Y))
            if d1 == 0:
                x, y = off2X, off2Y
            else:
                factor = d2 / d1
                x = on3X + (off3X - on3X) * factor
                y = on3Y + (off3Y - on3Y) * factor
            return (off1X, off1Y), (x, y), (on3X, on3Y)
        else:
            # just move the off curve point
            x = off2X + (off4X - on2X)
            y = off2Y + (off4Y - on2Y)
            return (off1X, off1Y), (x, y), (on3X, on3Y)

    if (off2X, off2Y) == (on2X, on2Y) == (off3X, off3Y) or (off2X, off2Y) == (on2X, on2Y) or (on2X, on2Y) == (off3X, off3Y):
        ## one or more bcps are on the joined point
        return (off1X, off1Y), (off4X, off4Y), (on3X, on3Y)

    if (on1X, on1Y) == (off1X, off1Y):
        off1X = off1X + (off2X - off1X) * .1
        off1Y = off1Y + (off2Y - off1Y) * .1

    if (on3X, on3Y) == (off4X, off4Y):
        off4X = off4X + (off3X - off4X) * .1
        off4Y = off4Y + (off3Y - off4Y) * .1

    smooth = False
    if (off2X, off2Y) != (on2X, on2Y) and (off3X, off3Y) != (on2X, on2Y):
        dx1, dy1 = on2X - off2X, on2Y - off2Y
        dx2, dy2 = off3X - on2X, off3Y - on2Y
        a1 = atan2(dx1, dy1)
        a2 = atan2(dx2, dy2)
        if abs(a1 - a2) < 0.05:
            smooth = True

    # first calculate an approximaly t
    d1 = _distance((on2X, on2Y), (off2X, off2Y))
    d2 = d1 + _distance((off3X, off3Y), (on2X, on2Y))

    if d2 == 0:
        t = 0
    else:
        t = d1 / d2

    # cut of the extreme t values
    error = .15
    if smooth:
        error = 0
    if t < error:
        t = error
    elif t > 1 - error:
        t = 1 - error

    # just multiply the first handle of the first curve by t
    p2X = on1X + (off1X - on1X) * (1 / t)
    p2Y = on1Y + (off1Y - on1Y) * (1 / t)
    # and the last handle of the last curve by t

    p3X = on3X + (off4X - on3X) * (1 / (1 - t))
    p3Y = on3Y + (off4Y - on3Y) * (1 / (1 - t))

    return (p2X, p2Y), (p3X, p3Y), (on3X, on3Y)

if __name__ == "__main__":
    import doctest
    doctest.testmod()



import weakref
from fontTools.ttLib import TTFont
from fontTools.feaLib.builder import addOpenTypeFeaturesFromString
from defcon.objects.base import BaseObject

try:
    import compositor
    from defcon.objects.uniData import UnicodeData
    from defcon.objects.features import Features
except ImportError:
    pass

# ---------
# Factories
# ---------

def _makeCMAP(unicodeData):
    mapping = {}
    for name, values in unicodeData.items():
        mapping[name] = values[0]
    return mapping

def _layoutEngineOTLTablesRepresentationFactory(layoutEngine):
    font = layoutEngine.font
    gdef = gsub = gpos = None
    if font.features.text:
        otf = TTFont()
        otf.setGlyphOrder(sorted(font.keys()))
        # compile with fontTools
        try:
            addOpenTypeFeaturesFromString(otf, font.features.text)
        except:
            import traceback
            print(traceback.format_exc(5))
        if "GDEF" in otf:
            gdef = otf["GDEF"]
        if "GSUB" in otf:
            gsub = otf["GSUB"]
        if "GPOS" in otf:
            gpos = otf["GPOS"]
    return gdef, gsub, gpos

# -----------
# Main Object
# -----------

class LayoutEngine(BaseObject):

    """
    This object provides a GDEF, GSUB and GPOS OpenType Layout Engine for
    the default layer of the given font. The engine uses the ``compositor``
    module so you must have that installed to use this object.

    **This object posts the following notifications:**

    - LayoutEngine.Changed

    This object monitors the font's feature text and character mapping. When
    those change, the compiled tables will be flagged for recompilation and
    the next time the engine is queried the tables will be recompiled. Any
    data that you have retrieved, such as the list of feature tags, may no
    longer be correct. Thus, the data will need to be retrieved again. To be
    notified when this is necessary, subscribe to the ``LayoutEngine.Changed``
    notification.
    """

    changeNotificationName = "LayoutEngine.Changed"
    representationFactories = {
        "defcon.layoutEngine.tables" : dict(
            factory=_layoutEngineOTLTablesRepresentationFactory,
            destructiveNotifications=("LayoutEngine._DestroyCachedTables")
        )
    }

    def __init__(self, font):
        self._needsInternalUpdate = True
        self._font = weakref.ref(font)
        self._layoutEngine = compositor.LayoutEngine()
        super(LayoutEngine, self).__init__()
        self.beginSelfNotificationObservation()

    def _get_font(self):
        if self._font is not None:
            return self._font()
        return None

    font = property(_get_font)

    def _get_engine(self):
        if self._needsInternalUpdate:
            self._updateEngine()
        return self._layoutEngine

    engine = property(_get_engine, doc="The compositor layout engine. This object must always be retrieved from the LayoutEngine for the automatic updating to occur.")

    # --------------
    # Engine Updates
    # --------------

    def _updateEngine(self):
        font = self.font
        cmap = _makeCMAP(font.unicodeData)
        self._layoutEngine.setCMAP(cmap)
        gdef, gsub, gpos = self.getRepresentation("defcon.layoutEngine.tables")
        self._layoutEngine.setFeatureTables(gdef, gsub, gpos)
        self._needsInternalUpdate = False

    # -------------
    # Notifications
    # -------------

    def beginSelfNotificationObservation(self):
        super(LayoutEngine, self).beginSelfNotificationObservation()
        self.beginSelfLayersObservation()
        self.beginSelfLayerObservation()
        self.beginSelfFeaturesObservation()

    def endSelfNotificationObservation(self):
        self.endSelfLayersObservation()
        self.endSelfLayerObservation()
        self.endSelfFeaturesObservation()
        super(LayoutEngine, self).endSelfNotificationObservation()
        self._font = None

    # default layer changed (changes cmap)

    def beginSelfLayersObservation(self):
        layers = self.font.layers
        layers.addObserver(observer=self, methodName="_layerSetDefaultLayerWillChange", notification="LayerSet.DefaultLayerWillChange")
        layers.addObserver(observer=self, methodName="_layerSetDefaultLayerChanged", notification="LayerSet.DefaultLayerChanged")

    def endSelfLayersObservation(self):
        layers = self.font.layers
        layers.removeObserver(observer=self, notification="LayerSet.DefaultLayerWillChange")
        layers.removeObserver(observer=self, notification="LayerSet.DefaultLayerChanged")

    def _layerSetDefaultLayerWillChange(self, notification):
        self.endSelfLayerObservation()

    def _layerSetDefaultLayerChanged(self, notification):
        self.beginLayerObservation()
        self._postNeedsUpdateNotification()

    # cmap change

    def beginSelfLayerObservation(self):
        layer = self.font.layers.defaultLayer
        layer.addObserver(observer=self, methodName="_layerGlyphUnicodesChanged", notification="Layer.GlyphUnicodesChanged")

    def endSelfLayerObservation(self):
        layer = self.font.layers.defaultLayer
        layer.removeObserver(observer=self, notification="Layer.GlyphUnicodesChanged")

    def _layerGlyphUnicodesChanged(self):
        self._postNeedsUpdateNotification()

    # feature text change

    def beginSelfFeaturesObservation(self):
        features = self.font.features
        features.addObserver(observer=self, methodName="_featuresTextChanged", notification="Features.TextChanged")

    def endSelfFeaturesObservation(self):
        features = self.font.features
        features.removeObserver(observer=self, notification="Features.TextChanged")

    def _featuresTextChanged(self, notification):
        self._destroyCachedTables()
        self._postNeedsUpdateNotification()

    # posting

    def _destroyCachedTables(self):
        self.postNotification("LayoutEngine._DestroyCachedTables")

    def _postNeedsUpdateNotification(self):
        self._needsInternalUpdate = True
        self.postNotification(self.changeNotificationName)

    # ----------
    # Engine API
    # ----------

    def process(self, stringOrGlyphList, script="latn", langSys=None, rightToLeft=False, case="unchanged"):
        """
        Process a string (or list of glyph names) with the current
        feature states for the given **script** and **langSys**.

        The writing will be left to right unless **rightToLeft**
        is set to True.

        The case may be changed following the Unicode case conversion
        rules by setting **case** to one of the following:

        +-----------+
        | unchanged |
        +-----------+
        | upper     |
        +-----------+
        | lower     |
        +-----------+
        """
        self._updateEngine()
        glyphRecords = self.engine.process(
                stringOrGlyphList,
                script=script, langSys=langSys,
                rightToLeft=rightToLeft, case=case
            )
        layer = self.font.layers.defaultLayer
        finalGlyphRecords = []
        for glyphRecord in glyphRecords:
            if glyphRecord.glyphName not in layer:
                continue
            layerGlyph = layer[glyphRecord.glyphName]
            glyphRecord.advanceWidth += layerGlyph.width
            glyphRecord.advanceHeight += layerGlyph.height
            finalGlyphRecords.append(glyphRecord)
        return finalGlyphRecords

    def getScriptList(self):
        """
        Get a list of defined scripts.
        """
        return self.engine.getScriptList()

    def getLanguageList(self):
        """
        Get a list of defined languages.
        """
        return self.engine.getLanguageList()

    def getFeatureList(self):
        """
        Get a list of defined features.
        """
        return self.engine.getFeatureList()

    def getFeatureState(self, name):
        """
        Get the state for the feature with **name**.
        """
        return self.engine.getFeatureState(name)

    def setFeatureState(self, name, state):
        """
        Set the state for the feature with **name**.
        """
        self.engine.setFeatureState(name, state)



import unittest
from defcon.tools.fuzzyNumber import FuzzyNumber


class TestFuzzyNumber(unittest.TestCase):

    def __init__(self, methodName):
        unittest.TestCase.__init__(self, methodName)

    def test_init(self):
        fuzzyNumber1 = FuzzyNumber(value=0, threshold=1)
        fuzzyNumber2 = FuzzyNumber(2, 3)
        self.assertEqual([fuzzyNumber1.value, fuzzyNumber1.threshold],
                         [0, 1])
        self.assertEqual([fuzzyNumber2.value, fuzzyNumber2.threshold],
                         [2, 3])

    def test_repr(self):
        fuzzyNumber = FuzzyNumber(0, 1)
        self.assertEqual(repr(fuzzyNumber), "[0.000000 1.000000]")

    def test_comparison(self):
        fuzzyNumber1 = FuzzyNumber(value=0, threshold=1)
        self.assertEqual(fuzzyNumber1, 0)
        self.assertTrue(fuzzyNumber1 < 1)
        self.assertFalse(fuzzyNumber1 < -0.000001)
        self.assertFalse(fuzzyNumber1 < 0)

        fuzzyNumber2 = FuzzyNumber(value=0.999999, threshold=1)
        self.assertEqual(
            repr(sorted([fuzzyNumber1, fuzzyNumber2])),
            "[[0.000000 1.000000], [0.999999 1.000000]]"
        )
        self.assertFalse(fuzzyNumber1 < fuzzyNumber2)

        fuzzyNumber2 = FuzzyNumber(value=1, threshold=1)
        self.assertEqual(
            repr(sorted([fuzzyNumber1, fuzzyNumber2])),
            "[[0.000000 1.000000], [1.000000 1.000000]]"
        )
        self.assertTrue(fuzzyNumber1 < fuzzyNumber2)

        fuzzyNumber2 = FuzzyNumber(value=-0.999999, threshold=1)
        self.assertEqual(
            repr(sorted([fuzzyNumber1, fuzzyNumber2])),
            "[[0.000000 1.000000], [-0.999999 1.000000]]"
        )
        self.assertFalse(fuzzyNumber1 > fuzzyNumber2)

        fuzzyNumber2 = FuzzyNumber(value=-1, threshold=1)
        self.assertEqual(
            repr(sorted([fuzzyNumber1, fuzzyNumber2])),
            "[[-1.000000 1.000000], [0.000000 1.000000]]"
        )
        self.assertTrue(fuzzyNumber1 > fuzzyNumber2)

        # equal
        self.assertEqual(fuzzyNumber1, fuzzyNumber1)
        self.assertNotEqual(fuzzyNumber1, fuzzyNumber2)

        # complex sorting
        fuzzyNumber2 = FuzzyNumber(value=0.999999, threshold=1)
        self.assertEqual(
            repr(sorted([(fuzzyNumber1, 20), (fuzzyNumber2, 10)])),
            "[([0.999999 1.000000], 10), ([0.000000 1.000000], 20)]"
        )


if __name__ == "__main__":
    unittest.main()



class FuzzyNumber(object):

    def __init__(self, value, threshold):
        self.value = value
        self.threshold = threshold

    def __repr__(self):
        return "[%f %f]" % (self.value, self.threshold)

    def __lt__(self, other):
        if hasattr(other, "value"):
            if abs(self.value - other.value) < self.threshold:
                return False
            else:
                return self.value < other.value
        return self.value < other

    def __eq__(self, other):
        if hasattr(other, "value"):
            return abs(self.value - other.value) < self.threshold
        return self.value == other

    def __hash__(self):
        return hash((self.value, self.threshold))



from __future__ import absolute_import
import weakref
from fontTools.ufoLib import UFOReader, UFOFileStructure, UFOFormatVersion
from defcon.objects.base import BaseObject
from defcon.objects.layer import Layer


class LayerSet(BaseObject):

    """
    This object manages all layers in the font.

    **This object posts the following notifications:**

    - LayerSet.Changed
    - LayerSet.LayersChanged
    - LayerSet.LayerChanged
    - LayerSet.DefaultLayerWillChange
    - LayerSet.DefaultLayerChanged
    - LayerSet.LayerOrderChanged
    - LayerSet.LayerAdded
    - LayerSet.LayerDeleted
    - LayerSet.LayerWillBeDeleted
    - LayerSet.LayerNameChanged

    This object behaves like a dict. For example, to get a particular
    layer::

        layer = layerSet["layer name"]

    If the layer name is None, the default layer will be retrieved.

    Note: It's up to the caller to ensure that a default layer is present
    as required by the UFO specification.
    """

    changeNotificationName = "LayerSet.Changed"
    representationFactories = {}

    def __init__(self, font=None, layerClass=None, libClass=None, unicodeDataClass=None,
            guidelineClass=None, glyphClass=None,
            glyphContourClass=None, glyphPointClass=None, glyphComponentClass=None, glyphAnchorClass=None,
            glyphImageClass=None):

        if font is not None:
            font = weakref.ref(font)
        self._font = font
        super(LayerSet, self).__init__()
        self.beginSelfNotificationObservation()

        if layerClass is None:
            layerClass = Layer
        self._layerClass = layerClass
        self._libClass = libClass
        self._unicodeDataClass = unicodeDataClass
        self._glyphClass = glyphClass
        self._glyphContourClass = glyphContourClass
        self._glyphPointClass = glyphPointClass
        self._glyphComponentClass = glyphComponentClass
        self._glyphAnchorClass = glyphAnchorClass
        self._glyphImageClass = glyphImageClass
        self._guidelineClass = guidelineClass

        self._layers = {}
        self._layerOrder = []
        self._defaultLayer = None

        self._layerActionHistory = []
        self._dirty = False

    def __del__(self):
        super(LayerSet, self).__del__()
        self._layers = None

    # --------------
    # Parent Objects
    # --------------

    def getParent(self):
        return self.font

    def _get_font(self):
        if self._font is None:
            return None
        return self._font()

    font = property(_get_font, doc="The :class:`Font` that this layer set belongs to.")

    # -------------
    # Default Layer
    # -------------

    def _get_defaultLayerName(self):
        defaultLayer = self.defaultLayer
        for name, layer in self._layers.items():
            if layer == defaultLayer:
                return name

    _defaultLayerName = property(_get_defaultLayerName)

    def _get_defaultLayer(self):
        return self._defaultLayer

    def _set_defaultLayer(self, layer):
        if layer is None:
            raise ValueError("The default layer must not be None.")
        if layer == self._defaultLayer:
            return
        self.postNotification(notification="LayerSet.DefaultLayerWillChange")
        oldName = None
        if self._defaultLayer is not None:
            oldName = self._defaultLayer.name
        self._defaultLayer = layer
        self._layerActionHistory.append(dict(action="default", newDefault=layer.name, oldDefault=oldName))
        self.postNotification(notification="LayerSet.DefaultLayerChanged", data=dict(oldValue=oldName, newValue=layer.name))
        self.dirty = True

    defaultLayer = property(_get_defaultLayer, _set_defaultLayer, doc="The default :class:`Layer` object. Setting this will post *LayerSet.DefaultLayerChanged* and *LayerSet.Changed* notifications.")

    # -----------
    # Layer Order
    # -----------

    def _get_layerOrder(self):
        return list(self._layerOrder)

    def _set_layerOrder(self, order):
        oldOrder = self._layerOrder
        if self._layerOrder == order:
            return
        assert len(order) == len(self._layerOrder)
        assert set(order) == set(self._layerOrder)
        self._layerOrder = list(order)
        self.postNotification(notification="LayerSet.LayerOrderChanged", data=dict(oldValue=oldOrder, newValue=order))
        self.dirty = True

    layerOrder = property(_get_layerOrder, _set_layerOrder, doc="The layer order from top to bottom. Setting this will post *LayerSet.LayerOrderChanged* and *LayerSet.Changed* notifications.")

    # -------------
    # Layer Creation
    # -------------

    def instantiateLayer(self, glyphSet):
        layer = self._layerClass(
            layerSet=self,
            glyphSet=glyphSet,
            libClass=self._libClass,
            unicodeDataClass=self._unicodeDataClass,
            glyphClass=self._glyphClass,
            glyphContourClass=self._glyphContourClass,
            glyphPointClass=self._glyphPointClass,
            glyphComponentClass=self._glyphComponentClass,
            glyphAnchorClass=self._glyphAnchorClass,
            guidelineClass=self._guidelineClass,
            glyphImageClass=self._glyphImageClass
        )
        return layer

    def beginSelfLayerNotificationObservation(self, layer):
        layer.addObserver(observer=self, methodName="_layerDirtyStateChange", notification="Layer.Changed")
        layer.addObserver(observer=self, methodName="_layerNameChange", notification="Layer.NameChanged")

    def endSelfLayerNotificationObservation(self, layer):
        if layer.dispatcher is None:
            return
        layer.removeObserver(observer=self, notification="Layer.Changed")
        layer.removeObserver(observer=self, notification="Layer.NameChanged")
        layer.endSelfNotificationObservation()

    def newLayer(self, name, glyphSet=None):
        """
        Create a new :class:`Layer` and add it to
        the top of the layer order. **glyphSet** should
        only be passed when reading from a UFO.

        This posts *LayerSet.LayerAdded* and *LayerSet.Changed* notifications.
        """
        if name in self._layers:
            raise KeyError("A layer named \"%s\" already exists." % name)
        assert name is not None
        layer = self.instantiateLayer(glyphSet)
        self.beginSelfLayerNotificationObservation(layer)
        layer.disableNotifications()
        layer.name = name
        if glyphSet is None:
            layer.dirty = True
        else:
            glyphSet.readLayerInfo(layer)
            layer.dirty = False
        layer.enableNotifications()
        self._stampLayerInfoDataState(layer)
        self._layers[name] = layer
        self._layerOrder.append(name)
        self._layerActionHistory.append(dict(action="new", name=name))
        self.postNotification("LayerSet.LayerAdded", data=dict(name=name))
        self.postNotification("LayerSet.LayersChanged")
        self.dirty = True
        return layer

    # -------------
    # Dict Behavior
    # -------------

    def __iter__(self):
        for name in self.layerOrder:
            yield self[name]

    def __getitem__(self, name):
        if name is None:
            name = self._defaultLayerName
        return self._layers[name]

    def __delitem__(self, name):
        if name is None:
            name = self._defaultLayerName
        if name not in self:
            raise KeyError("%s not in layers" % name)
        self.postNotification("LayerSet.LayerWillBeDeleted", data=dict(name=name))
        layer = self._layers[name]
        self.endSelfLayerNotificationObservation(layer)
        del self._layers[name]
        self._layerOrder.remove(name)
        self._layerActionHistory.append(dict(action="delete", name=name))
        self.postNotification("LayerSet.LayerDeleted", data=dict(name=name))
        self.postNotification("LayerSet.LayersChanged")
        self.dirty = True

    def __len__(self):
        return len(self.layerOrder)

    def __contains__(self, name):
        if name is None:
            name = self._defaultLayerName
        return name in self._layers

    # ----
    # Save
    # ----

    def getSaveProgressBarTickCount(self, formatVersion):
        """
        Get the number of ticks that will be used by a progress bar
        in the save method. This method should not be called externally.
        Subclasses may override this method to implement custom saving behavior.
        """
        count = 0
        if formatVersion < 3:
            count += 1
            count += self.defaultLayer.getSaveProgressBarTickCount(formatVersion)
        else:
            for layer in self:
                count += 1
                count += layer.getSaveProgressBarTickCount(formatVersion)
        return count

    def save(self, writer, saveAs=False, progressBar=None):
        """
        Save all layers. This method should not be called externally.
        Subclasses may override this method to implement custom saving behavior.
        """
        # work through the layer action history
        if not saveAs:
            for actionData in self._layerActionHistory:
                action = actionData["action"]
                if action == "delete":
                    layerName = actionData["name"]
                    if layerName in writer.layerContents:
                        writer.deleteGlyphSet(layerName)
                elif action == "rename":
                    oldName = actionData["oldName"]
                    newName = actionData["newName"]
                    if oldName in writer.layerContents:
                        writer.renameGlyphSet(oldName, newName)
                        if newName == self.defaultLayer.name:
                            writer.renameGlyphSet(newName, newName, defaultLayer=True)
                elif action == "default":
                    newDefault = actionData["newDefault"]
                    oldDefault = actionData["oldDefault"]
                    # if either one is already in the writer.layerContents it should be renamed
                    # otherwise it will be handled by the creation of the glyph set
                    if oldDefault in writer.layerContents:
                        # rename the old default layer by itself and unflag it as the default layer
                        writer.renameGlyphSet(oldDefault, oldDefault, defaultLayer=False)
                    if newDefault in writer.layerContents:
                        # rename the new default layer by itself and flag it as the default layer
                        writer.renameGlyphSet(newDefault, newDefault, defaultLayer=True)

                elif action == "new":
                    # this will be handled by the creation of the glyph set
                    pass
        # save the layers

        if writer.formatVersionTuple < UFOFormatVersion.FORMAT_3_0:
            if progressBar is not None:
                progressBar.update(text="Saving glyphs...", increment=0)
            layer = self.defaultLayer
            glyphSet = writer.getGlyphSet(layerName=None, defaultLayer=True, validateRead=self.ufoLibReadValidate, validateWrite=self.ufoLibWriteValidate)
            layer.save(glyphSet, saveAs=saveAs, progressBar=progressBar)
            if progressBar is not None:
                progressBar.update()
        else:
            for layerName in self.layerOrder:
                if progressBar is not None:
                    progressBar.update(text="Saving layer \"%s\"..." % layerName, increment=0)
                layer = self[layerName]
                isDefaultLayer = layer == self.defaultLayer
                glyphSet = writer.getGlyphSet(layerName=layerName, defaultLayer=isDefaultLayer, validateRead=self.ufoLibReadValidate, validateWrite=self.ufoLibWriteValidate)
                layer.save(glyphSet, saveAs=saveAs, progressBar=progressBar)
                # this prevents us from saving when the color was deleted
                #if layer.lib or layer.color:
                glyphSet.writeLayerInfo(layer)
                self._stampLayerInfoDataState(layer)
                layer.dirty = False
                if progressBar is not None:
                    progressBar.update()
            writer.writeLayerContents(self.layerOrder)
        # reset the action history
        self._layerActionHistory = []
        # if < UFO 3 was written, flag all of the non-default layers as "new"
        defaultLayer = self.defaultLayer
        for layer in self:
            if layer == defaultLayer:
                continue
            self._layerActionHistory.append(dict(action="new", name=layer.name))

    def _fontSaveWasCompleted(self):
        """
        On saving a UFO, the underlying fs object is closed.
        The objects then stored in layer._glyphSet contain references
        to a closed, and therefore unusable, filesystem. To remedy this,
        after the save is completed this method will be called and new
        GlyphSet objects will be created and assigned to the layers.
        """
        font = self.font
        font.close()
        font._reader = reader = UFOReader(font.path, validate=font.ufoLibReadValidate)
        defaultLayerName = self._defaultLayerName
        for layerName in self.layerOrder:
            layer = self[layerName]
            if reader.formatVersionTuple < (3, 0):
                # ufo2 or less has not layers
                # map only to the default layer to the ufo2 glyphSet
                # set all all other layers internal _glyphSet object to None
                if defaultLayerName == layerName:
                    glyphSet = reader.getGlyphSet(validateRead=self.ufoLibReadValidate)
                else:
                    glyphSet = None
            else:
                # we do have layers and the must be available after a save action
                glyphSet = reader.getGlyphSet(layerName=layerName, validateRead=self.ufoLibReadValidate)

            layer._glyphSet = glyphSet

    # ------------------------
    # Notification Observation
    # ------------------------

    def endSelfNotificationObservation(self):
        if self.dispatcher is None:
            return
        for layer in self._layers.values():
            self.endSelfLayerNotificationObservation(layer)
        super(LayerSet, self).endSelfNotificationObservation()
        self._font = None

    def _layerDirtyStateChange(self, notification):
        self.postNotification("LayerSet.LayerChanged")
        self.dirty = True

    def _layerNameChange(self, notification):
        data = notification.data
        oldName = data["oldName"]
        newName = data["newName"]
        self._layers[newName] = self._layers.pop(oldName)
        index = self._layerOrder.index(oldName)
        self._layerOrder.pop(index)
        self._layerOrder.insert(index, newName)
        self._layerActionHistory.append(dict(action="rename", oldName=oldName, newName=newName))
        self.postNotification("LayerSet.LayerNameChanged")

    # ---------------------
    # External Edit Support
    # ---------------------

    def _stampLayerInfoDataState(self, layer, glyphSet=None):
        if glyphSet is None:
            glyphSet = layer._glyphSet
        if glyphSet is None:
            return
        # there isn't a mod time function
        # so load the data and pack it.
        i = _StaticLayerInfoMaker()
        glyphSet.readLayerInfo(i)
        layer._dataOnDisk = i.pack()

    def testForExternalChanges(self, reader):
        """
        Test for external changes. This should not be called externally.
        """
        # changed default
        defaultLayerName = self._defaultLayerName
        onDiskDefaultLayerName = reader.getDefaultLayerName()
        defaultLayerChanged = defaultLayerName != onDiskDefaultLayerName
        # changed layer order
        onDiskLayerOrder = reader.getLayerNames()
        layerOrderChanged = onDiskLayerOrder != self.layerOrder
        # layers added since we started up
        addedLayers = []
        for layerName in set(onDiskLayerOrder) - set(self.layerOrder):
            # try to filter out layers that were removed in memory
            wasDeletedInMemory = False
            for actionData in self._layerActionHistory:
                action = actionData["action"]
                if action == "delete" and actionData["name"] == layerName:
                    wasDeletedInMemory = True
            if not wasDeletedInMemory:
                addedLayers.append(layerName)
        # layers deleted since we started up
        deletedLayers = list(set(self.layerOrder) - set(onDiskLayerOrder))
        # modified layers
        modifiedLayers = {}
        for layerName in self.layerOrder:
            if layerName in deletedLayers:
                continue
            layer = self[layerName]
            modifiedGlyphs, addedGlyphs, deletedGlyphs = layer.testForExternalChanges(reader)
            newLayerInfo = _StaticLayerInfoMaker()
            layerInfoChanged = False
            glyphSet = reader.getGlyphSet(layerName=layer.name, validateRead=layer.ufoLibReadValidate)
            if glyphSet is not None:
                glyphSet.readLayerInfo(newLayerInfo)
                layerInfoChanged = layer._dataOnDisk != newLayerInfo.pack()
            if modifiedGlyphs or addedGlyphs or deletedGlyphs or layerInfoChanged:
                modifiedLayers[layerName] = dict(
                    info=layerInfoChanged,
                    modified=modifiedGlyphs,
                    added=addedGlyphs,
                    deleted=deletedGlyphs
                )
        # pack
        result = dict(
            defaultLayer=defaultLayerChanged,
            order=layerOrderChanged,
            added=addedLayers,
            deleted=deletedLayers,
            modified=modifiedLayers
        )
        # cross your fingers
        return result

    def reloadLayers(self, layerData):
        """
        Reload the layers. This should not be called externally.
        """
        with UFOReader(self.font.path, validate=self.font.ufoLibReadValidate) as reader:
            # handle the layers
            currentLayerOrder = self.layerOrder
            for layerName, l in layerData.get("layers", {}).items():
                # new layer
                if layerName not in currentLayerOrder:
                    glyphSet = reader.getGlyphSet(layerName, validateRead=self.ufoLibReadValidate, validateWrite=self.font.ufoLibWriteValidate)
                    self.newLayer(layerName, glyphSet)
                # get the layer
                layer = self[layerName]
                # reload the layer info
                if l.get("info"):
                    layer.color = None
                    layer.lib.clear()
                    layer._glyphSet.readLayerInfo(layer)
                    self._stampLayerInfoDataState(layer)
                # reload the glyphs
                glyphNames = l.get("glyphNames", [])
                if glyphNames:
                    layer.reloadGlyphs(glyphNames)
            # handle the order
            if layerData.get("order", False):
                newLayerOrder = reader.getLayerNames()
                for layerName in self.layerOrder:
                    if layerName not in newLayerOrder:
                        newLayerOrder.append(layerName)
                self.layerOrder = newLayerOrder
            # handle the default layer
            if layerData.get("default", False):
                newDefaultLayerName = reader.getDefaultLayerName()
                self.defaultLayer = self[newDefaultLayerName]

    # -----------------------------
    # Serialization/Deserialization
    # -----------------------------

    def getDataForSerialization(self, **kwargs):
        serialize = lambda item: item.getDataForSerialization()

        def get_layers(k):
            layers = []
            for name in self.layerOrder:
                layer = self[name]
                isDefaultLayer = layer == self.defaultLayer
                layers.append((name, serialize(layer), isDefaultLayer))
            return layers

        getters = [('layers', get_layers)]

        return self._serialize(getters, **kwargs)

    def setDataFromSerialization(self, data):
        if 'layers' not in data:
            return
        for name, data, isDefault in data['layers']:
            layer = self.newLayer(name)
            layer.setDataFromSerialization(data)
            if isDefault:
                self.defaultLayer = layer

class _StaticLayerInfoMaker(object):

    def __init__(self):
        self.lib = {}
        self.color = None

    def pack(self):
        from fontTools.misc import plistlib
        data = {}
        if self.lib:
            data["lib"] = self.lib
        if self.color is not None:
            data["color"] = self.color
        return plistlib.dumps(data)


if __name__ == "__main__":
    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from fontTools import unicodedata

_openClosePairText = """
# This was automatically generated from UnicodeData.txt and then
# tweaked by hand to handle the special exceptions.

0028;LEFT PARENTHESIS;Ps
0029;RIGHT PARENTHESIS;Pe

005B;LEFT SQUARE BRACKET;Ps
005D;RIGHT SQUARE BRACKET;Pe

007B;LEFT CURLY BRACKET;Ps
007D;RIGHT CURLY BRACKET;Pe

00AB;LEFT-POINTING DOUBLE ANGLE QUOTATION MARK;Pi
00BB;RIGHT-POINTING DOUBLE ANGLE QUOTATION MARK;Pf

0F3A;TIBETAN MARK GUG RTAGS GYON;Ps
0F3B;TIBETAN MARK GUG RTAGS GYAS;Pe

0F3C;TIBETAN MARK ANG KHANG GYON;Ps
0F3D;TIBETAN MARK ANG KHANG GYAS;Pe

169B;OGHAM FEATHER MARK;Ps
169C;OGHAM REVERSED FEATHER MARK;Pe

2018;LEFT SINGLE QUOTATION MARK;Pi
2019;RIGHT SINGLE QUOTATION MARK;Pf

201C;LEFT DOUBLE QUOTATION MARK;Pi
201D;RIGHT DOUBLE QUOTATION MARK;Pf

2039;SINGLE LEFT-POINTING ANGLE QUOTATION MARK;Pi
203A;SINGLE RIGHT-POINTING ANGLE QUOTATION MARK;Pf

2045;LEFT SQUARE BRACKET WITH QUILL;Ps
2046;RIGHT SQUARE BRACKET WITH QUILL;Pe

207D;SUPERSCRIPT LEFT PARENTHESIS;Ps
207E;SUPERSCRIPT RIGHT PARENTHESIS;Pe

208D;SUBSCRIPT LEFT PARENTHESIS;Ps
208E;SUBSCRIPT RIGHT PARENTHESIS;Pe

2308;LEFT CEILING;Ps
2309;RIGHT CEILING;Pe

230A;LEFT FLOOR;Ps
230B;RIGHT FLOOR;Pe

2329;LEFT-POINTING ANGLE BRACKET;Ps
232A;RIGHT-POINTING ANGLE BRACKET;Pe

2768;MEDIUM LEFT PARENTHESIS ORNAMENT;Ps
2769;MEDIUM RIGHT PARENTHESIS ORNAMENT;Pe

276A;MEDIUM FLATTENED LEFT PARENTHESIS ORNAMENT;Ps
276B;MEDIUM FLATTENED RIGHT PARENTHESIS ORNAMENT;Pe

276C;MEDIUM LEFT-POINTING ANGLE BRACKET ORNAMENT;Ps
276D;MEDIUM RIGHT-POINTING ANGLE BRACKET ORNAMENT;Pe

276E;HEAVY LEFT-POINTING ANGLE QUOTATION MARK ORNAMENT;Ps
276F;HEAVY RIGHT-POINTING ANGLE QUOTATION MARK ORNAMENT;Pe

2770;HEAVY LEFT-POINTING ANGLE BRACKET ORNAMENT;Ps
2771;HEAVY RIGHT-POINTING ANGLE BRACKET ORNAMENT;Pe

2772;LIGHT LEFT TORTOISE SHELL BRACKET ORNAMENT;Ps
2773;LIGHT RIGHT TORTOISE SHELL BRACKET ORNAMENT;Pe

2774;MEDIUM LEFT CURLY BRACKET ORNAMENT;Ps
2775;MEDIUM RIGHT CURLY BRACKET ORNAMENT;Pe

27C5;LEFT S-SHAPED BAG DELIMITER;Ps
27C6;RIGHT S-SHAPED BAG DELIMITER;Pe

27E6;MATHEMATICAL LEFT WHITE SQUARE BRACKET;Ps
27E7;MATHEMATICAL RIGHT WHITE SQUARE BRACKET;Pe

27E8;MATHEMATICAL LEFT ANGLE BRACKET;Ps
27E9;MATHEMATICAL RIGHT ANGLE BRACKET;Pe

27EA;MATHEMATICAL LEFT DOUBLE ANGLE BRACKET;Ps
27EB;MATHEMATICAL RIGHT DOUBLE ANGLE BRACKET;Pe

27EC;MATHEMATICAL LEFT WHITE TORTOISE SHELL BRACKET;Ps
27ED;MATHEMATICAL RIGHT WHITE TORTOISE SHELL BRACKET;Pe

27EE;MATHEMATICAL LEFT FLATTENED PARENTHESIS;Ps
27EF;MATHEMATICAL RIGHT FLATTENED PARENTHESIS;Pe

2983;LEFT WHITE CURLY BRACKET;Ps
2984;RIGHT WHITE CURLY BRACKET;Pe

2985;LEFT WHITE PARENTHESIS;Ps
2986;RIGHT WHITE PARENTHESIS;Pe

2987;Z NOTATION LEFT IMAGE BRACKET;Ps
2988;Z NOTATION RIGHT IMAGE BRACKET;Pe

2989;Z NOTATION LEFT BINDING BRACKET;Ps
298A;Z NOTATION RIGHT BINDING BRACKET;Pe

298B;LEFT SQUARE BRACKET WITH UNDERBAR;Ps
298C;RIGHT SQUARE BRACKET WITH UNDERBAR;Pe

298D;LEFT SQUARE BRACKET WITH TICK IN TOP CORNER;Ps
298E;RIGHT SQUARE BRACKET WITH TICK IN BOTTOM CORNER;Pe

298F;LEFT SQUARE BRACKET WITH TICK IN BOTTOM CORNER;Ps
2990;RIGHT SQUARE BRACKET WITH TICK IN TOP CORNER;Pe

2991;LEFT ANGLE BRACKET WITH DOT;Ps
2992;RIGHT ANGLE BRACKET WITH DOT;Pe

2993;LEFT ARC LESS-THAN BRACKET;Ps
2994;RIGHT ARC GREATER-THAN BRACKET;Pe

2995;DOUBLE LEFT ARC GREATER-THAN BRACKET;Ps
2996;DOUBLE RIGHT ARC LESS-THAN BRACKET;Pe

2997;LEFT BLACK TORTOISE SHELL BRACKET;Ps
2998;RIGHT BLACK TORTOISE SHELL BRACKET;Pe

29D8;LEFT WIGGLY FENCE;Ps
29D9;RIGHT WIGGLY FENCE;Pe

29DA;LEFT DOUBLE WIGGLY FENCE;Ps
29DB;RIGHT DOUBLE WIGGLY FENCE;Pe

29FC;LEFT-POINTING CURVED ANGLE BRACKET;Ps
29FD;RIGHT-POINTING CURVED ANGLE BRACKET;Pe

2E02;LEFT SUBSTITUTION BRACKET;Pi
2E03;RIGHT SUBSTITUTION BRACKET;Pf

2E04;LEFT DOTTED SUBSTITUTION BRACKET;Pi
2E05;RIGHT DOTTED SUBSTITUTION BRACKET;Pf

2E09;LEFT TRANSPOSITION BRACKET;Pi
2E0A;RIGHT TRANSPOSITION BRACKET;Pf

2E0C;LEFT RAISED OMISSION BRACKET;Pi
2E0D;RIGHT RAISED OMISSION BRACKET;Pf

2E1C;LEFT LOW PARAPHRASE BRACKET;Pi
2E1D;RIGHT LOW PARAPHRASE BRACKET;Pf

2E20;LEFT VERTICAL BAR WITH QUILL;Pi
2E21;RIGHT VERTICAL BAR WITH QUILL;Pf

2E22;TOP LEFT HALF BRACKET;Ps
2E23;TOP RIGHT HALF BRACKET;Pe

2E24;BOTTOM LEFT HALF BRACKET;Ps
2E25;BOTTOM RIGHT HALF BRACKET;Pe

2E26;LEFT SIDEWAYS U BRACKET;Ps
2E27;RIGHT SIDEWAYS U BRACKET;Pe

2E28;LEFT DOUBLE PARENTHESIS;Ps
2E29;RIGHT DOUBLE PARENTHESIS;Pe

3008;LEFT ANGLE BRACKET;Ps
3009;RIGHT ANGLE BRACKET;Pe

300A;LEFT DOUBLE ANGLE BRACKET;Ps
300B;RIGHT DOUBLE ANGLE BRACKET;Pe

300C;LEFT CORNER BRACKET;Ps
300D;RIGHT CORNER BRACKET;Pe

300E;LEFT WHITE CORNER BRACKET;Ps
300F;RIGHT WHITE CORNER BRACKET;Pe

3010;LEFT BLACK LENTICULAR BRACKET;Ps
3011;RIGHT BLACK LENTICULAR BRACKET;Pe

3014;LEFT TORTOISE SHELL BRACKET;Ps
3015;RIGHT TORTOISE SHELL BRACKET;Pe

3016;LEFT WHITE LENTICULAR BRACKET;Ps
3017;RIGHT WHITE LENTICULAR BRACKET;Pe

3018;LEFT WHITE TORTOISE SHELL BRACKET;Ps
3019;RIGHT WHITE TORTOISE SHELL BRACKET;Pe

301A;LEFT WHITE SQUARE BRACKET;Ps
301B;RIGHT WHITE SQUARE BRACKET;Pe

301D;REVERSED DOUBLE PRIME QUOTATION MARK;Ps
301E;DOUBLE PRIME QUOTATION MARK;Pe

FD3F;ORNATE RIGHT PARENTHESIS;Ps
FD3E;ORNATE LEFT PARENTHESIS;Pe

FE17;PRESENTATION FORM FOR VERTICAL LEFT WHITE LENTICULAR BRACKET;Ps
FE18;PRESENTATION FORM FOR VERTICAL RIGHT WHITE LENTICULAR BRAKCET;Pe

FE35;PRESENTATION FORM FOR VERTICAL LEFT PARENTHESIS;Ps
FE36;PRESENTATION FORM FOR VERTICAL RIGHT PARENTHESIS;Pe

FE37;PRESENTATION FORM FOR VERTICAL LEFT CURLY BRACKET;Ps
FE38;PRESENTATION FORM FOR VERTICAL RIGHT CURLY BRACKET;Pe

FE39;PRESENTATION FORM FOR VERTICAL LEFT TORTOISE SHELL BRACKET;Ps
FE3A;PRESENTATION FORM FOR VERTICAL RIGHT TORTOISE SHELL BRACKET;Pe

FE3B;PRESENTATION FORM FOR VERTICAL LEFT BLACK LENTICULAR BRACKET;Ps
FE3C;PRESENTATION FORM FOR VERTICAL RIGHT BLACK LENTICULAR BRACKET;Pe

FE3D;PRESENTATION FORM FOR VERTICAL LEFT DOUBLE ANGLE BRACKET;Ps
FE3E;PRESENTATION FORM FOR VERTICAL RIGHT DOUBLE ANGLE BRACKET;Pe

FE3F;PRESENTATION FORM FOR VERTICAL LEFT ANGLE BRACKET;Ps
FE40;PRESENTATION FORM FOR VERTICAL RIGHT ANGLE BRACKET;Pe

FE41;PRESENTATION FORM FOR VERTICAL LEFT CORNER BRACKET;Ps
FE42;PRESENTATION FORM FOR VERTICAL RIGHT CORNER BRACKET;Pe

FE43;PRESENTATION FORM FOR VERTICAL LEFT WHITE CORNER BRACKET;Ps
FE44;PRESENTATION FORM FOR VERTICAL RIGHT WHITE CORNER BRACKET;Pe

FE47;PRESENTATION FORM FOR VERTICAL LEFT SQUARE BRACKET;Ps
FE48;PRESENTATION FORM FOR VERTICAL RIGHT SQUARE BRACKET;Pe

FE59;SMALL LEFT PARENTHESIS;Ps
FE5A;SMALL RIGHT PARENTHESIS;Pe

FE5B;SMALL LEFT CURLY BRACKET;Ps
FE5C;SMALL RIGHT CURLY BRACKET;Pe

FE5D;SMALL LEFT TORTOISE SHELL BRACKET;Ps
FE5E;SMALL RIGHT TORTOISE SHELL BRACKET;Pe

FF08;FULLWIDTH LEFT PARENTHESIS;Ps
FF09;FULLWIDTH RIGHT PARENTHESIS;Pe

FF3B;FULLWIDTH LEFT SQUARE BRACKET;Ps
FF3D;FULLWIDTH RIGHT SQUARE BRACKET;Pe

FF5B;FULLWIDTH LEFT CURLY BRACKET;Ps
FF5D;FULLWIDTH RIGHT CURLY BRACKET;Pe

FF5F;FULLWIDTH LEFT WHITE PARENTHESIS;Ps
FF60;FULLWIDTH RIGHT WHITE PARENTHESIS;Pe

FF62;HALFWIDTH LEFT CORNER BRACKET;Ps
FF63;HALFWIDTH RIGHT CORNER BRACKET;Pe

# Special Exceptions
# These have flipped relatives that are already mapped above.

201F;DOUBLE HIGH-REVERSED-9 QUOTATION MARK;Pi
201D;RIGHT DOUBLE QUOTATION MARK;Pf # already mapped

201E;DOUBLE LOW-9 QUOTATION MARK;Ps
201D;RIGHT DOUBLE QUOTATION MARK;Pf # already mapped

201B;SINGLE HIGH-REVERSED-9 QUOTATION MARK;Pi
2019;RIGHT SINGLE QUOTATION MARK;Pf # already mapped

201A;SINGLE LOW-9 QUOTATION MARK;Ps
2019;RIGHT SINGLE QUOTATION MARK;Pf # already mapped

2E42;DOUBLE LOW-REVERSED-9 QUOTATION MARK;Ps
201F;DOUBLE HIGH-REVERSED-9 QUOTATION MARK;Pi # already mapped

301D;REVERSED DOUBLE PRIME QUOTATION MARK;Ps  # already mapped
301F;LOW DOUBLE PRIME QUOTATION MARK;Pe
"""

# load the data

_openToClose = {}
_closeToOpen = {}

openValue = None
for line in _openClosePairText.splitlines():
    line = line.split("#")[0].strip()
    if not line:
        continue
    value, name, category = line.split(";")
    value = int(value, 16)
    if openValue is not None:
        closeValue = value
        if openValue not in _openToClose:
            _openToClose[openValue] = closeValue
        if closeValue not in _closeToOpen:
            _closeToOpen[closeValue] = openValue
        openValue = None
    else:
        openValue = value

# ordered sets
orderedScripts = []
for value in unicodedata.Scripts.VALUES:
    value = unicodedata.script_name(value)
    if value == "Unknown":
        continue
    if value not in orderedScripts:
        orderedScripts.append(value)
orderedScripts.append("Unknown")

orderedBlocks = []
for value in unicodedata.Blocks.VALUES:
    if value == "No_Block":
        continue
    if value not in orderedBlocks:
        orderedBlocks.append(value)

orderedCategories = """Lu
Ll
Lt
Lm
Lo
Mn
Mc
Me
Nd
Nl
No
Pc
Pd
Ps
Pe
Pi
Pf
Po
Sm
Sc
Sk
So
Zs
Zl
Zp
Cc
Cf
Cs
Co
Cn""".splitlines()

# functions

def decompositionBase(value):
    letterCategories = ("Ll", "Lu", "Lt", "Lo")
    try:
        c = chr(value)
    # see not in category function
    except ValueError:
        return -1
    decomposition = unicodedata.decomposition(c)
    if decomposition.startswith("<"):
        return -1
    if " " not in decomposition:
        return -1
    parts = decomposition.split(" ")
    unichrs = [chr(int(i, 16)) for i in parts if i]
    letters = [ord(i) for i in unichrs if unicodedata.category(i) in letterCategories]
    letterCount = len(letters)
    if letterCount != 1:
        return -1
    decomposedUniValue = letters[0]
    furtherDecomposedUniValue = decompositionBase(decomposedUniValue)
    if furtherDecomposedUniValue != -1:
        furtherFurtherDecomposedUniValue = decompositionBase(furtherDecomposedUniValue)
        if furtherFurtherDecomposedUniValue != -1:
            decomposedUniValue = furtherFurtherDecomposedUniValue
        else:
            decomposedUniValue = furtherDecomposedUniValue
    return decomposedUniValue

def openRelative(value):
    return _closeToOpen.get(value)

def closeRelative(value):
    return _openToClose.get(value)

def category(value):
    c = chr(value)
    return unicodedata.category(c)

def script(value):
    char = chr(value)
    return unicodedata.script_name(unicodedata.script(char), default="Unknown")

def block(value):
    char = chr(value)
    return unicodedata.block(char)



from __future__ import absolute_import
import weakref
from warnings import warn
from fontTools.misc.transform import Transform
from defcon.objects.base import BaseObject
from defcon.tools.identifiers import makeRandomIdentifier
from defcon.tools.representations import componentBoundsRepresentationFactory, componentPointBoundsRepresentationFactory

_defaultTransformation = (1, 0, 0, 1, 0, 0)


class Component(BaseObject):

    """
    This object represents a reference to another glyph.

    **This object posts the following notifications:**

    - Component.Changed
    - Component.BaseGlyphChanged
    - Component.BaseGlyphDataChanged
    - Component.TransformationChanged
    - Component.IdentifierChanged

    """

    changeNotificationName = "Component.Changed"
    representationFactories = {
        "defcon.component.bounds" : dict(
            factory=componentBoundsRepresentationFactory,
            destructiveNotifications=("Component.TransformationChanged", "Component.BaseGlyphChanged")
        ),
        "defcon.component.controlPointBounds" : dict(
            factory=componentPointBoundsRepresentationFactory,
            destructiveNotifications=("Component.TransformationChanged", "Component.BaseGlyphChanged")
        )
    }

    def __init__(self, glyph=None):
        self._font = None
        self._layerSet = None
        self._layer = None
        self._glyph = None
        self.glyph = glyph

        self._dirty = False
        self._baseGlyph = None
        self._transformation = tuple(_defaultTransformation)
        self._identifier = None

        super(Component, self).__init__()
        self.beginSelfNotificationObservation()

    # ----------
    # Attributes
    # ----------

    # parents

    def getParent(self):
        return self.glyph

    def _get_font(self):
        font = None
        if self._font is None:
            glyph = self.glyph
            if glyph is not None:
                font = glyph.font
                if font is not None:
                    self._font = weakref.ref(font)
        else:
            font = self._font()
        return font

    font = property(_get_font, doc="The :class:`Font` that this component belongs to.")

    def _get_layerSet(self):
        layerSet = None
        if self._layerSet is None:
            glyph = self.glyph
            if glyph is not None:
                layerSet = glyph.layerSet
                if layerSet is not None:
                    self._layerSet = weakref.ref(layerSet)
        else:
            layerSet = self._layerSet()
        return layerSet

    layerSet = property(_get_layerSet, doc="The :class:`LayerSet` that this component belongs to.")

    def _get_layer(self):
        layer = None
        if self._layer is None:
            glyph = self.glyph
            if glyph is not None:
                layer = glyph.layer
                if layer is not None:
                    self._layer = weakref.ref(layer)
        else:
            layer = self._layer()
        return layer

    layer = property(_get_layer, doc="The :class:`Layer` that this component belongs to.")

    def _get_glyph(self):
        if self._glyph is None:
            return None
        return self._glyph()

    def _set_glyph(self, glyph):
        assert self._glyph is None
        if glyph is not None:
            glyph = weakref.ref(glyph)
        self._font = None
        self._layerSet = None
        self._layer = None
        self._glyph = glyph

    glyph = property(_get_glyph, _set_glyph, doc="The :class:`Glyph` that this component belongs to. This should not be set externally.")

    # bounds

    def _get_bounds(self):
        return self.getRepresentation("defcon.component.bounds")

    bounds = property(_get_bounds, doc="The bounds of the components's outline expressed as a tuple of form (xMin, yMin, xMax, yMax).")

    def _get_controlPointBounds(self):
        return self.getRepresentation("defcon.component.controlPointBounds")

    controlPointBounds = property(_get_controlPointBounds, doc="The control bounds of all points in the components. This only measures the point positions, it does not measure curves. So, curves without points at the extrema will not be properly measured.")

    # base glyph

    def _set_baseGlyph(self, value):
        newBaseGlyph = value
        oldBaseGlyph = self._baseGlyph
        if newBaseGlyph == oldBaseGlyph:
            return
        self.endSelfBaseGlyphNotificationObservation()
        self._baseGlyph = newBaseGlyph
        self.beginSelfBaseGlyphNotificationObservation()
        self.postNotification(notification="Component.BaseGlyphChanged", data=dict(oldValue=oldBaseGlyph, newValue=newBaseGlyph))
        self.dirty = True

    def _get_baseGlyph(self):
        return self._baseGlyph

    baseGlyph = property(_get_baseGlyph, _set_baseGlyph, doc="The glyph that the components references. Setting this will post *Component.BaseGlyphChanged* and *Component.Changed* notifications.")

    # transformation

    def _set_transformation(self, value):
        oldValue = self._transformation
        if value == oldValue:
            return
        self._transformation = value
        self.postNotification(notification="Component.TransformationChanged", data=dict(oldValue=oldValue, newValue=value))
        self.dirty = True

    def _get_transformation(self):
        return self._transformation

    transformation = property(_get_transformation, _set_transformation, doc="The transformation matrix for the component. Setting this will post *Component.TransformationChanged* and *Component.Changed* notifications.")

    # -----------
    # Pen Methods
    # -----------

    def draw(self, pen):
        """
        Draw the component with **pen**.
        """
        from fontTools.pens.pointPen import PointToSegmentPen
        pointPen = PointToSegmentPen(pen)
        self.drawPoints(pointPen)

    def drawPoints(self, pointPen):
        """
        Draw the component with **pointPen**.
        """
        try:
            pointPen.addComponent(self._baseGlyph, self._transformation, identifier=self.identifier)
        except TypeError:
            pointPen.addComponent(self._baseGlyph, self._transformation)
            warn("The addComponent method needs an identifier kwarg. The component's identifier value has been discarded.", DeprecationWarning)

    # ----
    # Move
    # ----

    def move(self, values):
        """
        Move the component by **(x, y)**.

        This posts *Component.TransformationChanged* and *Component.Changed* notifications.
        """
        (x, y) = values
        xScale, xyScale, yxScale, yScale, xOffset, yOffset = self._transformation
        xOffset += x
        yOffset += y
        self.transformation = (xScale, xyScale, yxScale, yScale, xOffset, yOffset)

    # ------------
    # Point Inside
    # ------------

    def pointInside(self, coordinates, evenOdd=False):
        """
        Returns a boolean indicating if **(x, y)** is in the
        "black" area of the component.
        """
        (x, y) = coordinates
        from fontTools.pens.pointInsidePen import PointInsidePen
        glyph = self.glyph
        if glyph is None:
            return False
        font = self.font
        if font is None:
            return False
        pen = PointInsidePen(glyphSet=font, testPoint=(x, y), evenOdd=evenOdd)
        self.draw(pen)
        return pen.getResult()

    # ----------
    # Identifier
    # ----------

    def _get_identifiers(self):
        identifiers = None
        glyph = self.glyph
        if glyph is not None:
            identifiers = glyph.identifiers
        if identifiers is None:
            identifiers = set()
        return identifiers

    identifiers = property(_get_identifiers, doc="Set of identifiers for the glyph that this component belongs to. This is primarily for internal use.")

    def _get_identifier(self):
        return self._identifier

    def _set_identifier(self, value):
        # don't allow overwritting an existing identifier
        if self._identifier is not None:
            return
        oldIdentifier = self.identifier
        if value == oldIdentifier:
            return
        # don't allow a duplicate
        identifiers = self.identifiers
        assert value not in identifiers
        # free the old identifier
        if oldIdentifier in identifiers:
            identifiers.remove(oldIdentifier)
        # store
        self._identifier = value
        if value is not None:
            identifiers.add(value)
        # post notifications
        self.postNotification("Component.IdentifierChanged", data=dict(oldValue=oldIdentifier, newValue=value))
        self.dirty = True

    identifier = property(_get_identifier, _set_identifier, doc="The identifier. Setting this will post *Component.IdentifierChanged* and *Component.Changed* notifications.")

    def generateIdentifier(self):
        """
        Create a new, unique identifier for and assign it to the contour.
        This will post *Component.IdentifierChanged* and *Component.Changed* notifications.
        """
        if self.identifier is None:
            identifier = makeRandomIdentifier(existing=self.identifiers)
            self.identifier = identifier
        return self.identifier

    # ------------------------
    # Notification Observation
    # ------------------------

    def beginSelfNotificationObservation(self):
        super(Component, self).beginSelfNotificationObservation()
        self.beginSelfBaseGlyphNotificationObservation()

    def endSelfNotificationObservation(self):
        self.endSelfBaseGlyphNotificationObservation()
        super(Component, self).endSelfNotificationObservation()
        self._font = None
        self._layerSet = None
        self._layer = None
        self._glyph = None

    def beginSelfBaseGlyphNotificationObservation(self):
        baseGlyph = self.baseGlyph
        if baseGlyph is None:
            return
        dispatcher = self.dispatcher
        if dispatcher is None:
            return
        layer = self.layer
        # base glyph is available
        if baseGlyph in layer:
            self._beginBaseGlyphObservations()
        # base glyph is not available
        else:
            self._beginLayerObservations()

    def endSelfBaseGlyphNotificationObservation(self):
        dispatcher = self.dispatcher
        if dispatcher is None:
            return
        self._endBaseGlyphObservations()
        self._endLayerObservations()

    def _beginBaseGlyphObservations(self, baseGlyph=None):
        layer = self.layer
        if baseGlyph is None:
            baseGlyph = layer[self.baseGlyph]
        baseGlyph.addObserver(self, "baseGlyphNameChangedNotificationCallback", "Glyph.NameChanged")
        baseGlyph.addObserver(self, "baseGlyphDataChangedNotificationCallback", "Glyph.ContoursChanged")
        baseGlyph.addObserver(self, "baseGlyphDataChangedNotificationCallback", "Glyph.ComponentsChanged")
        layer.addObserver(self, "layerGlyphWillBeDeletedNotificationCallback", "Layer.GlyphWillBeDeleted")

    def _endBaseGlyphObservations(self, baseGlyph=None):
        layer = self.layer
        if baseGlyph is None:
            baseGlyph = self.baseGlyph
            if baseGlyph is None:
                return
            if baseGlyph in layer:
                baseGlyph = layer[baseGlyph]
            else:
                return
        if baseGlyph.hasObserver(self, "Glyph.NameChanged"):
            baseGlyph.removeObserver(self, "Glyph.NameChanged")
            baseGlyph.removeObserver(self, "Glyph.ContoursChanged")
            baseGlyph.removeObserver(self, "Glyph.ComponentsChanged")
        if layer.hasObserver(self, "Layer.GlyphWillBeDeleted"):
            layer.removeObserver(self, "Layer.GlyphWillBeDeleted")

    def _beginLayerObservations(self):
        layer = self.layer
        layer.addObserver(self, "layerGlyphNameChangedNotificationCallback", "Layer.GlyphNameChanged")
        layer.addObserver(self, "layerGlyphAddedNotificationCallback", "Layer.GlyphAdded")

    def _endLayerObservations(self):
        layer = self.layer
        if not layer.hasObserver(self, "Layer.GlyphNameChanged"):
            return
        layer.removeObserver(self, "Layer.GlyphNameChanged")
        layer.removeObserver(self, "Layer.GlyphAdded")

    def baseGlyphNameChangedNotificationCallback(self, notification):
        newName = notification.data["newValue"]
        layer = self.layer
        notBaseGlyph = layer[newName]
        self._endBaseGlyphObservations(notBaseGlyph)
        self._beginLayerObservations()

    def layerGlyphNameChangedNotificationCallback(self, notification):
        newName = notification.data["newValue"]
        baseGlyph = self.baseGlyph
        if newName != baseGlyph:
            return
        self._endLayerObservations()
        self._beginBaseGlyphObservations()

    def layerGlyphWillBeDeletedNotificationCallback(self, notification):
        name = notification.data["name"]
        if name != self.baseGlyph:
            return
        self._endBaseGlyphObservations()
        self._beginLayerObservations()

    def layerGlyphAddedNotificationCallback(self, notification):
        name = notification.data["name"]
        if name != self.baseGlyph:
            return
        self._endLayerObservations()
        self._beginBaseGlyphObservations()

    def baseGlyphDataChangedNotificationCallback(self, notification):
        self.postNotification("Component.BaseGlyphDataChanged")

    # -----------------------------
    # Serialization/Deserialization
    # -----------------------------

    def getDataForSerialization(self, **kwargs):
        from functools import partial

        simple_get = partial(getattr, self)
        getters = (
            ('baseGlyph', simple_get),
            ('transformation', simple_get),
            ('identifier', simple_get)
        )

        return self._serialize(getters, **kwargs)

    def setDataFromSerialization(self, data):
        from functools import partial

        simple_set = partial(setattr, self)
        setters = (
            ('baseGlyph', simple_set),
            ('transformation', simple_set),
            ('identifier', simple_set)
        )
        for key, setter in setters:
            if key not in data:
                continue
            setter(key, data[key])


if __name__ == "__main__":
    import doctest
    doctest.testmod()



"""
PointPen for reversing the winding direction of contours.

NOTE: The module is deprecated and the ``ReverseContourPointPen`` class has
been moved to ``fontTools.pens.pointPen`` module.
"""

from fontTools.pens.pointPen import AbstractPointPen, ReverseContourPointPen
import warnings


warnings.warn(
    "Importing the `defcon.pens.reverseContourPointPen` module is deprecated. "
    "Use `from fontTools.pens.pointPen import ReverseContourPointPen` instead.",
    DeprecationWarning, stacklevel=2)
